# Copyright (c) 2011-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Algorithms related to discretization of continuous dynamics.

See Also
========
L{find_controller}
"""
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import os
import warnings
import pprint
from copy import deepcopy
import multiprocessing as mp

import numpy as np
from scipy import sparse as sp
import polytope as pc

from polytope.plot import plot_partition, plot_transition_arrow
from tulip import transys as trs
from tulip.hybrid import LtiSysDyn, PwaSysDyn

from .prop2partition import (PropPreservingPartition,
                             pwa_partition, part2convex)
from .feasible import is_feasible, solve_feasible
from .plot import plot_ts_on_partition

# inline imports:
#
# inline: import matplotlib.pyplot as plt

debug = False

class AbstractSwitched(object):
    """Abstraction of SwitchedSysDyn, with mode-specific and common info.

    Attributes:

      - ppp: merged partition, if any
          Preserves both propositions and dynamics

      - ts: common TS, if any

      - ppp2ts: map from C{ppp.regions} to C{ts.states}

      - modes: dict of {mode: AbstractPwa}

      - ppp2modes: map from C{ppp.regions} to C{modes[mode].ppp.regions}
        of the form:

        {mode: list}

        where C{list} has same indices as C{ppp.regions} and
        elements in each C{list} are indices of regions in
        each C{modes[mode].ppp.regions}.

        type: dict

    Each partition corresponds to some mode.
    (for switched systems)

    In each mode a L{PwaSysDyn} is active.
    """
    def __init__(
        self, ppp=None, ts=None, ppp2ts=None,
        modes=None, ppp2modes=None
    ):
        if modes is None:
            modes = dict()

        self.ppp = ppp
        self.ts = ts
        self.ppp2ts = ppp2ts
        self.modes = modes
        self.ppp2modes = ppp2modes

    def __str__(self):
        s = 'Abstraction of switched system\n'
        s += str('common PPP:\n') + str(self.ppp)
        s += str('common ts:\n') + str(self.ts)

        for mode, ab in self.modes.iteritems():
            s += 'mode: ' + str(mode)
            s += ', with abstraction:\n' + str(ab)

        return s

    def ppp2pwa(self, mode, i):
        """Return original C{Region} containing C{Region} C{i} in C{mode}.

        @param mode: key of C{modes}

        @param i: Region index in common partition C{ppp.regions}.

        @return: tuple C{(j, region)} of:

                - index C{j} of C{Region} and
                - C{Region} object

            in C{modes[mode].ppp.regions}
        """
        region_idx = self.ppp2modes[mode][i]
        ab = self.modes[mode]
        return ab.ppp2pwa(region_idx)

    def ppp2sys(self, mode, i):
        """Return index of active PWA subsystem in C{mode},

        @param mode: key of C{modes}

        @param i: Region index in common partition C{ppp.regions}.

        @return: tuple C{(j, subsystem)} of:

                - index C{j} of PWA C{subsystem}
                - L{LtiSysDyn} object C{subsystem}
        """
        region_idx = self.ppp2modes[mode][i]
        ab = self.modes[mode]
        return ab.ppp2sys(region_idx)

    def plot(self, show_ts=False, only_adjacent=False):
        """Plot mode partitions and merged partition, if one exists.

        For details see L{AbstractPwa.plot}.
        """
        axs = []
        color_seed = 0

        # merged partition exists ?
        if self.ppp is not None:
            for mode in self.modes:
                env_mode, sys_mode = mode
                edge_label = {'env_actions':env_mode,
                              'sys_actions':sys_mode}

                ax = _plot_abstraction(
                    self, show_ts=False, only_adjacent=False,
                    color_seed=color_seed
                )
                plot_ts_on_partition(
                    self.ppp, self.ts, self.ppp2ts,
                    edge_label, only_adjacent, ax
                )
                axs += [ax]

        # plot mode partitions
        for mode, ab in self.modes.iteritems():
            ax = ab.plot(show_ts, only_adjacent, color_seed)
            ax.set_title('Abstraction for mode: ' + str(mode))
            axs += [ax]

        #if isinstance(self.ts, dict):
        #    for ts in self.ts:
        #        ax = ts.plot()
        #        axs += [ax]
        return axs

class AbstractPwa(object):
    """Discrete abstraction of PWA dynamics, with attributes:

      - ppp: Partition into Regions.
          Each Region corresponds to
          a discrete state of the abstraction

          type: L{PropPreservingPartition}

      - ts: Finite transition system abstracting the continuous system.
          Each state corresponds to a Region in C{ppp.regions}.
          It can be fed into discrete synthesis algorithms.

          type: L{FTS}

      - ppp2ts: bijection between C{ppp.regions} and C{ts.states}.
          Has common indices with C{ppp.regions}.
          Elements are states in C{ts.states}.
          (usually each state is a str)

          type: list of states

      - pwa: system dynamics

          type: L{PwaSysDyn}

      - pwa_ppp: partition preserving both:

            - propositions and
            - domains of PWA subsystems

          Used for non-conservative planning.
          If just L{LtiSysDyn}, then the only difference
          of C{pwa_ppp} from C{orig_ppp} is convexification.

          type: L{PropPreservingPartition}

      - orig_ppp: partition preserving only propositions
          i.e., agnostic of dynamics

          type: L{PropPreservingPartition}

      - disc_params: parameters used in discretization that
          should be passed to the controller refinement
          to ensure consistency

          type: dict

    If any of the above is not given,
    then it is initialized to None.

    Notes
    =====
      1. There could be some redundancy in ppp and ofts,
         in that they are both decorated with propositions.
         This might be useful to keep each of
         them as functional units on their own
         (possible to change later).

      2. The 'Pwa' in L{AbstractPwa} includes L{LtiSysDyn}
         as a special case.
    """
    def __init__(
        self, ppp=None, ts=None, ppp2ts=None,
        pwa=None, pwa_ppp=None, ppp2pwa=None, ppp2sys=None,
        orig_ppp=None, ppp2orig=None,
        disc_params=None
    ):
        if disc_params is None:
            disc_params = dict()

        self.ppp = ppp
        self.ts = ts
        self.ppp2ts = ppp2ts

        self.pwa = pwa
        self.pwa_ppp = pwa_ppp
        self._ppp2pwa = ppp2pwa
        self._ppp2sys = ppp2sys

        self.orig_ppp = orig_ppp
        self._ppp2orig = ppp2orig

        # original_regions -> pwa_ppp
        # ppp2orig -> ppp2pwa_ppp
        # ppp2pwa -> ppp2pwa_sys

        self.disc_params = disc_params

    def __str__(self):
        s = str(self.ppp)
        s += str(self.ts)

        s += 30 * '-' + '\n'

        s += 'Map PPP Regions ---> TS states:\n'
        s += self._ppp2other_str(self.ppp2ts) + '\n'

        s += 'Map PPP Regions ---> PWA PPP Regions:\n'
        s += self._ppp2other_str(self._ppp2pwa) + '\n'

        s += 'Map PPP Regions ---> PWA Subsystems:\n'
        s += self._ppp2other_str(self._ppp2sys) + '\n'

        s += 'Map PPP Regions ---> Original PPP Regions:\n'
        s += self._ppp2other_str(self._ppp2orig) + '\n'

        s += 'Discretization Options:\n\t'
        s += pprint.pformat(self.disc_params) +'\n'

        return s

    def ts2ppp(self, state):
        region_index = self.ppp2ts.index(state)
        region = self.ppp[region_index]
        return (region_index, region)

    def ppp2trans(self, region_index):
        """Return the transition set constraint and active subsystem,

        for non-conservative planning.
        """
        reg_idx, pwa_region = self.ppp2pwa(region_index)
        sys_idx, sys = self.ppp2sys(region_index)
        return pwa_region, sys

    def ppp2pwa(self, region_index):
        """Return dynamics and predicate-preserving region
        and its index for PWA subsystem active in given region.

        The returned region is the C{trans_set} used for
        non-conservative planning.

        @param region_index: index in C{ppp.regions}.

        @rtype: C{(i, pwa.pwa_ppp[i])}
        """
        j = self._ppp2pwa[region_index]
        pwa_region = self.pwa_ppp[j]
        return (j, pwa_region)

    def ppp2sys(self, region_index):
        """Return index and PWA subsystem active in indexed region.

        Semantics: j-th sub-system is active in i-th Region,
        where C{j = ppp2pwa[i]}

        @param region_index: index in C{ppp.regions}.

        @rtype: C{(i, pwa.list_subsys[i])}
        """
        # LtiSysDyn ?
        if self._ppp2sys is None:
            return (0, self.pwa)

        subsystem_idx = self._ppp2sys[region_index]
        subsystem = self.pwa.list_subsys[subsystem_idx]
        return (subsystem_idx, subsystem)

    def ppp2orig(self, region_index):
        """Return index and region of original partition.

        The original partition is w/o any dynamics,
        not even the PWA domains, only the polytopic predicates.

        @param region_index: index in C{ppp.regions}.

        @rtype: C{(i, orig_ppp.regions[i])}
        """
        j = self._ppp2orig[region_index]
        orig_region = self.orig_ppp[j]
        return (j, orig_region)

    def _ppp2other_str(self, ppp2other):
        if ppp2other is None:
            return ''

        s = ''
        for i, other in enumerate(ppp2other):
            s += '\t\t' + str(i) + ' -> ' + str(other) + '\n'
        return s

    def _debug_str_(self):
        s = str(self.ppp)
        s += str(self.ts)

        s += '(PWA + Prop)-Preserving Partition'
        s += str(self.pwa_ppp)

        s += 'Original Prop-Preserving Partition'
        s += str(self.orig_ppp)
        return s

    def plot(self, show_ts=False, only_adjacent=False,
             color_seed=None):
        """Plot partition and optionally feasible transitions.

        @param show_ts: plot feasible transitions on partition
        @type show_ts: bool

        @param only_adjacent: plot feasible transitions only
            between adjacent regions. This reduces clutter,
            but if horizon > 1 and not all horizon used,
            then some transitions could be hidden.
        @param only_adjacent: bool
        """
        ax = _plot_abstraction(self, show_ts, only_adjacent,
                               color_seed)
        return ax

    def verify_transitions(self):
        logger.info('verifying transitions...')

        for from_state, to_state in self.ts.transitions():
            i, from_region = self.ts2ppp(from_state)
            j, to_region = self.ts2ppp(to_state)

            trans_set, sys = self.ppp2trans(i)

            params = {'N', 'close_loop', 'use_all_horizon'}
            disc_params = {k:v for k,v in self.disc_params.iteritems()
                           if k in params}

            s0 = solve_feasible(from_region, to_region, sys,
                                trans_set=trans_set, **disc_params)

            msg = str(i) + ' ---> ' + str(j)

            if not from_region <= s0:
                logger.error('incorrect transition: ' + msg)

                isect = from_region.intersect(s0)
                ratio = isect.volume /from_region.volume
                logger.error('intersection volume: ' + str(ratio) + ' %')
            else:
                logger.info('correct transition: ' + msg)

def _plot_abstraction(ab, show_ts, only_adjacent, color_seed):
    if ab.ppp is None or ab.ts is None:
        warnings.warn('Either ppp or ts is None.')
        return

    if show_ts:
        ts = ab.ts
        ppp2ts = ab.ppp2ts
    else:
        ts = None
        ppp2ts = None

    ax = ab.ppp.plot(
        ts, ppp2ts, only_adjacent=only_adjacent,
        color_seed=color_seed
    )

    #ax = self.ts.plot()

    return ax

def discretize(
    part, ssys, N=10, min_cell_volume=0.1,
    closed_loop=True, conservative=False,
    max_num_poly=5, use_all_horizon=False,
    trans_length=1, remove_trans=False,
    abs_tol=1e-7,
    plotit=False, save_img=False, cont_props=None,
    plot_every=1
):
    """Refine the partition and establish transitions
    based on reachability analysis.

    Reference
    =========
    U{[NOTM12]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#notm12>}

    See Also
    ========
    L{prop2partition.pwa_partition}, L{prop2partition.part2convex}

    @param part: L{PropPreservingPartition} object
    @param ssys: L{LtiSysDyn} or L{PwaSysDyn} object
    @param N: horizon length
    @param min_cell_volume: the minimum volume of cells in the resulting
        partition.
    @param closed_loop: boolean indicating whether the `closed loop`
        algorithm should be used. default True.
    @param conservative: if true, force sequence in reachability analysis
        to stay inside starting cell. If false, safety
        is ensured by keeping the sequence inside a convexified
        version of the original proposition preserving cell.
    @param max_num_poly: maximum number of polytopes in a region to use in
        reachability analysis.
    @param use_all_horizon: in closed loop algorithm: if we should look
        for reachability also in less than N steps.
    @param trans_length: the number of polytopes allowed to cross in a
        transition.  a value of 1 checks transitions
        only between neighbors, a value of 2 checks
        neighbors of neighbors and so on.
    @param remove_trans: if True, remove found transitions between
        non-neighbors.
    @param abs_tol: maximum volume for an "empty" polytope

    @param plotit: plot partitioning as it evolves
    @type plotit: boolean,
        default = False

    @param save_img: save snapshots of partitioning to PDF files,
        requires plotit=True
    @type save_img: boolean,
        default = False

    @param cont_props: continuous propositions to plot
    @type cont_props: list of C{Polytope}

    @rtype: L{AbstractPwa}
    """
    start_time = os.times()[0]

    orig_ppp = part
    min_cell_volume = (min_cell_volume /np.finfo(np.double).eps
        *np.finfo(np.double).eps)

    ispwa = isinstance(ssys, PwaSysDyn)
    islti = isinstance(ssys, LtiSysDyn)

    if ispwa:
        (part, ppp2pwa, part2orig) = pwa_partition(ssys, part)
    else:
        part2orig = range(len(part))

    # Save original polytopes, require them to be convex
    if conservative:
        orig_list = None
        orig = [0]
    else:
        (part, new2old) = part2convex(part) # convexify
        part2orig = [part2orig[i] for i in new2old]

        # map new regions to pwa subsystems
        if ispwa:
            ppp2pwa = [ppp2pwa[i] for i in new2old]

        remove_trans = False # already allowed in nonconservative
        orig_list = []
        for poly in part:
            if len(poly) == 0:
                orig_list.append(poly.copy())
            elif len(poly) == 1:
                orig_list.append(poly[0].copy())
            else:
                raise Exception("discretize: "
                    "problem in convexification")
        orig = range(len(orig_list))

    # Cheby radius of disturbance set
    # (defined within the loop for pwa systems)
    if islti:
        if len(ssys.E) > 0:
            rd = ssys.Wset.chebR
        else:
            rd = 0.

    # Initialize matrix for pairs to check
    IJ = part.adj.copy()
    IJ = IJ.todense()
    IJ = np.array(IJ)
    logger.debug("\n Starting IJ: \n" + str(IJ) )

    # next line omitted in discretize_overlap
    IJ = reachable_within(trans_length, IJ,
                          np.array(part.adj.todense()) )

    # Initialize output
    num_regions = len(part)
    transitions = np.zeros(
        [num_regions, num_regions],
        dtype = int
    )
    sol = deepcopy(part.regions)
    adj = part.adj.copy()
    adj = adj.todense()
    adj = np.array(adj)

    # next 2 lines omitted in discretize_overlap
    if ispwa:
        subsys_list = list(ppp2pwa)
    else:
        subsys_list = None
    ss = ssys

    # init graphics
    if plotit:
        try:
            import matplotlib.pyplot as plt

            plt.ion()
            fig, (ax1, ax2) = plt.subplots(1, 2)
            ax1.axis('scaled')
            ax2.axis('scaled')
            file_extension = 'pdf'
        except:
            logger.error('failed to import matplotlib')
            plt = None
    else:
    	plt = None

    iter_count = 0

    # List of how many "new" regions
    # have been created for each region
    # and a list of original number of neighbors
    #num_new_reg = np.zeros(len(orig_list))
    #num_orig_neigh = np.sum(adj, axis=1).flatten() - 1

    progress = list()

    # Do the abstraction
    while np.sum(IJ) > 0:
        ind = np.nonzero(IJ)
        # i,j swapped in discretize_overlap
        i = ind[1][0]
        j = ind[0][0]
        IJ[j, i] = 0
        si = sol[i]
        sj = sol[j]

        si_tmp = deepcopy(si)
        sj_tmp = deepcopy(sj)

        #num_new_reg[i] += 1
        #print(num_new_reg)

        if ispwa:
            ss = ssys.list_subsys[subsys_list[i]]
            if len(ss.E) > 0:
                rd, xd = pc.cheby_ball(ss.Wset)
            else:
                rd = 0.

        if conservative:
            # Don't use trans_set
            trans_set = None
        else:
            # Use original cell as trans_set
            trans_set = orig_list[orig[i]]

        S0 = solve_feasible(
            si, sj, ss, N, closed_loop,
            use_all_horizon, trans_set, max_num_poly
        )

        msg = '\n Working with partition cells: ' + str(i) + ', ' + str(j)
        logger.info(msg)

        msg = '\t' + str(i) +' (#polytopes = ' +str(len(si) ) +'), and:\n'
        msg += '\t' + str(j) +' (#polytopes = ' +str(len(sj) ) +')\n'

        if ispwa:
            msg += '\t with active subsystem: '
            msg += str(subsys_list[i]) + '\n'

        msg += '\t Computed reachable set S0 with volume: '
        msg += str(S0.volume) + '\n'

        logger.debug(msg)

        #logger.debug('si \cap s0')
        isect = si.intersect(S0)
        vol1 = isect.volume
        risect, xi = pc.cheby_ball(isect)

        #logger.debug('si \ s0')
        diff = si.diff(S0)
        vol2 = diff.volume
        rdiff, xd = pc.cheby_ball(diff)

        # if pc.is_fulldim(pc.Region([isect]).intersect(diff)):
        #     logging.getLogger('tulip.polytope').setLevel(logging.DEBUG)
        #     diff = pc.mldivide(si, S0, save=True)
        #
        #     ax = S0.plot()
        #     ax.axis([0.0, 1.0, 0.0, 2.0])
        #     ax.figure.savefig('./img/s0.pdf')
        #
        #     ax = si.plot()
        #     ax.axis([0.0, 1.0, 0.0, 2.0])
        #     ax.figure.savefig('./img/si.pdf')
        #
        #     ax = isect.plot()
        #     ax.axis([0.0, 1.0, 0.0, 2.0])
        #     ax.figure.savefig('./img/isect.pdf')
        #
        #     ax = diff.plot()
        #     ax.axis([0.0, 1.0, 0.0, 2.0])
        #     ax.figure.savefig('./img/diff.pdf')
        #
        #     ax = isect.intersect(diff).plot()
        #     ax.axis([0.0, 1.0, 0.0, 2.0])
        #     ax.figure.savefig('./img/diff_cap_isect.pdf')
        #
        #     logger.error('Intersection \cap Difference != \emptyset')
        #
        #     assert(False)

        if vol1 <= min_cell_volume:
            logger.warning('\t too small: si \cap Pre(sj), ' +
                           'so discard intersection')
        if vol1 <= min_cell_volume and isect:
            logger.warning('\t discarded non-empty intersection: ' +
                           'consider reducing min_cell_volume')
        if vol2 <= min_cell_volume:
            logger.warning('\t too small: si \ Pre(sj), so not reached it')

        # We don't want our partitions to be smaller than the disturbance set
        # Could be a problem since cheby radius is calculated for smallest
        # convex polytope, so if we have a region we might throw away a good
        # cell.
        if (vol1 > min_cell_volume) and (risect > rd) and \
           (vol2 > min_cell_volume) and (rdiff > rd):

            # Make sure new areas are Regions and add proposition lists
            if len(isect) == 0:
                isect = pc.Region([isect], si.props)
            else:
                isect.props = si.props.copy()

            if len(diff) == 0:
                diff = pc.Region([diff], si.props)
            else:
                diff.props = si.props.copy()

            # replace si by intersection (single state)
            isect_list = pc.separate(isect)
            sol[i] = isect_list[0]

            # cut difference into connected pieces
            difflist = pc.separate(diff)

            difflist += isect_list[1:]
            n_isect = len(isect_list) -1

            num_new = len(difflist)

            # add each piece, as a new state
            for region in difflist:
                sol.append(region)

                # keep track of PWA subsystems map to new states
                if ispwa:
                    subsys_list.append(subsys_list[i])
            n_cells = len(sol)
            new_idx = xrange(n_cells-1, n_cells-num_new-1, -1)

            """Update transition matrix"""
            transitions = np.pad(transitions, (0,num_new), 'constant')

            transitions[i, :] = np.zeros(n_cells)
            for r in new_idx:
                #transitions[:, r] = transitions[:, i]
                # All sets reachable from start are reachable from both part's
                # except possibly the new part
                transitions[i, r] = 0
                transitions[j, r] = 0

            # sol[j] is reachable from intersection of sol[i] and S0
            if i != j:
                transitions[j, i] = 1

                # sol[j] is reachable from each piece os S0 \cap sol[i]
                #for k in xrange(n_cells-n_isect-2, n_cells):
                #    transitions[j, k] = 1

            """Update adjacency matrix"""
            old_adj = np.nonzero(adj[i, :])[0]

            # reset new adjacencies
            adj[i, :] = np.zeros([n_cells -num_new])
            adj[:, i] = np.zeros([n_cells -num_new])
            adj[i, i] = 1

            adj = np.pad(adj, (0, num_new), 'constant')

            for r in new_idx:
                adj[i, r] = 1
                adj[r, i] = 1
                adj[r, r] = 1

                if not conservative:
                    orig = np.hstack([orig, orig[i]])

            # adjacencies between pieces of isect and diff
            for r in new_idx:
                for k in new_idx:
                    if r is k:
                        continue

                    if pc.is_adjacent(sol[r], sol[k]):
                        adj[r, k] = 1
                        adj[k, r] = 1

            msg = ''
            if logger.getEffectiveLevel() <= logging.DEBUG:
                msg += '\t\n Adding states ' + str(i) + ' and '
                for r in new_idx:
                    msg += str(r) + ' and '
                msg += '\n'
                logger.debug(msg)

            for k in np.setdiff1d(old_adj, [i,n_cells-1]):
                # Every "old" neighbor must be the neighbor
                # of at least one of the new
                if pc.is_adjacent(sol[i], sol[k]):
                    adj[i, k] = 1
                    adj[k, i] = 1
                elif remove_trans and (trans_length == 1):
                    # Actively remove transitions between non-neighbors
                    transitions[i, k] = 0
                    transitions[k, i] = 0

                for r in new_idx:
                    if pc.is_adjacent(sol[r], sol[k]):
                        adj[r, k] = 1
                        adj[k, r] = 1
                    elif remove_trans and (trans_length == 1):
                        # Actively remove transitions between non-neighbors
                        transitions[r, k] = 0
                        transitions[k, r] = 0

            """Update IJ matrix"""
            IJ = np.pad(IJ, (0,num_new), 'constant')
            adj_k = reachable_within(trans_length, adj, adj)
            sym_adj_change(IJ, adj_k, transitions, i)

            for r in new_idx:
                sym_adj_change(IJ, adj_k, transitions, r)

            if logger.getEffectiveLevel() <= logging.DEBUG:
                msg = '\n\n Updated adj: \n' + str(adj)
                msg += '\n\n Updated trans: \n' + str(transitions)
                msg += '\n\n Updated IJ: \n' + str(IJ)
                logger.debug(msg)

            logger.info('Divided region: ' + str(i) + '\n')
        elif vol2 < abs_tol:
            logger.info('Found: ' + str(i) + ' ---> ' + str(j) + '\n')
            transitions[j,i] = 1
        else:
            if logger.level <= logging.DEBUG:
                msg = '\t Unreachable: ' + str(i) + ' --X--> ' + str(j) + '\n'
                msg += '\t\t diff vol: ' + str(vol2) + '\n'
                msg += '\t\t intersect vol: ' + str(vol1) + '\n'
                logger.debug(msg)
            else:
                logger.info('\t unreachable\n')
            transitions[j,i] = 0

        # check to avoid overlapping Regions
        if debug:
            tmp_part = PropPreservingPartition(
                domain=part.domain,
                regions=sol, adj=sp.lil_matrix(adj),
                prop_regions=part.prop_regions
            )
            assert(tmp_part.is_partition() )

        n_cells = len(sol)
        progress_ratio = 1 - float(np.sum(IJ) ) /n_cells**2
        progress += [progress_ratio]

        msg = '\t total # polytopes: ' + str(n_cells) + '\n'
        msg += '\t progress ratio: ' + str(progress_ratio) + '\n'
        logger.info(msg)

        iter_count += 1

        # no plotting ?
        if not plotit:
            continue
        if plt is None or plot_partition is None:
            continue
        if iter_count % plot_every != 0:
            continue

        tmp_part = PropPreservingPartition(
            domain=part.domain,
            regions=sol, adj=sp.lil_matrix(adj),
            prop_regions=part.prop_regions
        )

        # plot pair under reachability check
        ax2.clear()
        si_tmp.plot(ax=ax2, color='green')
        sj_tmp.plot(ax2, color='red', hatch='o', alpha=0.5)
        plot_transition_arrow(si_tmp, sj_tmp, ax2)

        S0.plot(ax2, color='none', hatch='/', alpha=0.3)
        fig.canvas.draw()

        # plot partition
        ax1.clear()
        plot_partition(tmp_part, transitions.T, ax=ax1, color_seed=23)

        # plot dynamics
        ssys.plot(ax1, show_domain=False)

        # plot hatched continuous propositions
        part.plot_props(ax1)

        fig.canvas.draw()

        # scale view based on domain,
        # not only the current polytopes si, sj
        l,u = part.domain.bounding_box
        ax2.set_xlim(l[0,0], u[0,0])
        ax2.set_ylim(l[1,0], u[1,0])

        if save_img:
            fname = 'movie' +str(iter_count).zfill(3)
            fname += '.' + file_extension
            fig.savefig(fname, dpi=250)
        plt.pause(1)

    new_part = PropPreservingPartition(
        domain=part.domain,
        regions=sol, adj=sp.lil_matrix(adj),
        prop_regions=part.prop_regions
    )

    # check completeness of adjacency matrix
    if debug:
        tmp_part = deepcopy(new_part)
        tmp_part.compute_adj()

    # Generate transition system and add transitions
    ofts = trs.FTS()

    adj = sp.lil_matrix(transitions.T)
    n = adj.shape[0]
    ofts_states = range(n)

    ofts.states.add_from(ofts_states)

    ofts.transitions.add_adj(adj, ofts_states)

    # Decorate TS with state labels
    atomic_propositions = set(part.prop_regions)
    ofts.atomic_propositions.add_from(atomic_propositions)
    for state, region in zip(ofts_states, sol):
        state_prop = region.props.copy()
        ofts.states.add(state, ap=state_prop)

    param = {
        'N':N,
        'trans_length':trans_length,
        'closed_loop':closed_loop,
        'conservative':conservative,
        'use_all_horizon':use_all_horizon,
        'min_cell_volume':min_cell_volume,
        'max_num_poly':max_num_poly
    }

    ppp2orig = [part2orig[x] for x in orig]

    end_time = os.times()[0]
    msg = 'Total abstraction time: ' +\
          str(end_time - start_time) + '[sec]'
    print(msg)
    logger.info(msg)

    if save_img and plt is not None:
        fig, ax = plt.subplots(1, 1)
        plt.plot(progress)
        ax.set_xlabel('iteration')
        ax.set_ylabel('progress ratio')
        ax.figure.savefig('progress.pdf')

    return AbstractPwa(
        ppp=new_part,
        ts=ofts,
        ppp2ts=ofts_states,
        pwa=ssys,
        pwa_ppp=part,
        ppp2pwa=orig,
        ppp2sys=subsys_list,
        orig_ppp=orig_ppp,
        ppp2orig=ppp2orig,
        disc_params=param
    )

def reachable_within(trans_length, adj_k, adj):
    """Find cells reachable within trans_length hops.
    """
    if trans_length <= 1:
        return adj_k

    k = 1
    while k < trans_length:
        adj_k = np.dot(adj_k, adj)
        k += 1
    adj_k = (adj_k > 0).astype(int)

    return adj_k

def sym_adj_change(IJ, adj_k, transitions, i):
    horizontal = adj_k[i, :] -transitions[i, :] > 0
    vertical = adj_k[:, i] -transitions[:, i] > 0

    IJ[i, :] = horizontal.astype(int)
    IJ[:, i] = vertical.astype(int)

# DEFUNCT until further notice
def discretize_overlap(closed_loop=False, conservative=False):
    """default False.

    UNDER DEVELOPMENT; function signature may change without notice.
    Calling will result in NotImplementedError.
    """
    raise NotImplementedError
#
#         if rdiff < abs_tol:
#             logger.info("Transition found")
#             transitions[i,j] = 1
#
#         elif (vol1 > min_cell_volume) & (risect > rd) & \
#                 (num_new_reg[i] <= num_orig_neigh[i]+1):
#
#             # Make sure new cell is Region and add proposition lists
#             if len(isect) == 0:
#                 isect = pc.Region([isect], si.props)
#             else:
#                 isect.props = si.props.copy()
#
#             # Add new state
#             sol.append(isect)
#             size = len(sol)
#
#             # Add transitions
#             transitions = np.hstack([transitions, np.zeros([size - 1, 1],
#                                     dtype=int) ])
#             transitions = np.vstack([transitions, np.zeros([1, size],
#                                     dtype=int) ])
#
#             # All sets reachable from orig cell are reachable from both cells
#             transitions[size-1,:] = transitions[i,:]
#             transitions[size-1,j] = 1   # j is reachable from new cell
#
#             # Take care of adjacency
#             old_adj = np.nonzero(adj[i,:])[0]
#
#             adj = np.hstack([adj, np.zeros([size - 1, 1], dtype=int) ])
#             adj = np.vstack([adj, np.zeros([1, size], dtype=int) ])
#             adj[i,size-1] = 1
#             adj[size-1,i] = 1
#             adj[size-1,size-1] = 1
#
#             for k in np.setdiff1d(old_adj,[i,size-1]):
#                 if pc.is_adjacent(sol[size-1],sol[k],overlap=True):
#                     adj[size-1,k] = 1
#                     adj[k, size-1] = 1
#                 else:
#                     # Actively remove (valid) transitions between non-neighbors
#                     transitions[size-1,k] = 0
#                     transitions[k,size-1] = 0
#
#             # Assign original proposition cell to new state and update counts
#             if not conservative:
#                 orig = np.hstack([orig, orig[i]])
#             print(num_new_reg)
#             num_new_reg = np.hstack([num_new_reg, 0])
#             num_orig_neigh = np.hstack([num_orig_neigh, np.sum(adj[size-1,:])-1])
#
#             logger.info("\n Adding state " + str(size-1) + "\n")
#
#             # Just add adjacent cells for checking,
#             # unless transition already found
#             IJ = np.hstack([IJ, np.zeros([size - 1, 1], dtype=int) ])
#             IJ = np.vstack([IJ, np.zeros([1, size], dtype=int) ])
#             horiz2 = adj[size-1,:] - transitions[size-1,:] > 0
#             verti2 = adj[:,size-1] - transitions[:,size-1] > 0
#             IJ[size-1,:] = horiz2.astype(int)
#             IJ[:,size-1] = verti2.astype(int)
#         else:
#             logger.info("No transition found, intersect vol: " + str(vol1) )
#             transitions[i,j] = 0
#
#     new_part = PropPreservingPartition(
#                    domain=part.domain,
#                    regions=sol, adj=np.array([]),
#                    trans=transitions, prop_regions=part.prop_regions,
#                    original_regions=orig_list, orig=orig)
#     return new_part

def multiproc_discretize(q, mode, ppp, cont_dyn, disc_params):
    global logger
    logger = mp.log_to_stderr()

    name = mp.current_process().name
    print('Abstracting mode: ' + str(mode) + ', on: ' + str(name))

    absys = discretize(ppp, cont_dyn, **disc_params)

    q.put((mode, absys))
    print('Worker: ' + str(name) + 'finished.')

def multiproc_get_transitions(
    q, absys, mode, ssys, params
):
    global logger
    logger = mp.log_to_stderr()

    name = mp.current_process().name
    print('Merged transitions for mode: ' + str(mode) + ', on: ' + str(name))

    trans = get_transitions(absys, mode, ssys, **params)

    q.put((mode, trans))
    print('Worker: ' + str(name) + 'finished.')

def multiproc_discretize_switched(
    ppp, hybrid_sys, disc_params=None,
    plot=False, show_ts=False, only_adjacent=True
):
    """Parallel implementation of discretize_switched.

    Uses the multiprocessing package.
    """
    logger.info('parallel discretize_switched started')

    modes = hybrid_sys.modes
    mode_nums = hybrid_sys.disc_domain_size

    q = mp.Queue()

    mode_args = dict()
    for mode in modes:
        cont_dyn = hybrid_sys.dynamics[mode]
        mode_args[mode] = (q, mode, ppp, cont_dyn, disc_params[mode])

    jobs = [mp.Process(target=multiproc_discretize, args=args)
            for args in mode_args.itervalues()]
    for job in jobs:
        job.start()

    # flush before join:
    #   http://stackoverflow.com/questions/19071529/
    abstractions = dict()
    for job in jobs:
        mode, absys = q.get()
        abstractions[mode] = absys

    for job in jobs:
        job.join()

    # merge their domains
    (merged_abstr, ap_labeling) = merge_partitions(abstractions)
    n = len(merged_abstr.ppp)
    logger.info('Merged partition has: ' + str(n) + ', states')

    # find feasible transitions over merged partition
    for mode in modes:
        cont_dyn = hybrid_sys.dynamics[mode]
        params = disc_params[mode]

        mode_args[mode] = (q, merged_abstr, mode, cont_dyn, params)

    jobs = [mp.Process(target=multiproc_get_transitions, args=args)
            for args in mode_args.itervalues()]

    for job in jobs:
        job.start()

    trans = dict()
    for job in jobs:
        mode, t = q.get()
        trans[mode] = t

    # merge the abstractions, creating a common TS
    merge_abstractions(merged_abstr, trans,
                       abstractions, modes, mode_nums)

    if plot:
        plot_mode_partitions(merged_abstr, show_ts, only_adjacent)

    return merged_abstr

def discretize_switched(
    ppp, hybrid_sys, disc_params=None,
    plot=False, show_ts=False, only_adjacent=True
):
    """Abstract switched dynamics over given partition.

    @type ppp: L{PropPreservingPartition}

    @param hybrid_sys: dynamics of switching modes
    @type hybrid_sys: L{SwitchedSysDyn}

    @param disc_params: discretization parameters passed to L{discretize} for
		each mode. See L{discretize} for details.
    @type disc_params: dict (keyed by mode) of dicts.

    @param plot: save partition images
    @type plot: bool

    @param show_ts, only_adjacent: options for L{AbstractPwa.plot}.

    @return: abstracted dynamics,
        some attributes are dict keyed by mode
    @rtype: L{AbstractSwitched}
    """
    if disc_params is None:
        disc_params = {'N':1, 'trans_length':1}

    logger.info('discretizing hybrid system')

    modes = hybrid_sys.modes
    mode_nums = hybrid_sys.disc_domain_size

    # discretize each abstraction separately
    abstractions = dict()
    for mode in modes:
        logger.debug(30*'-'+'\n')
        logger.info('Abstracting mode: ' + str(mode))

        cont_dyn = hybrid_sys.dynamics[mode]

        absys = discretize(
            ppp, cont_dyn,
            **disc_params[mode]
        )
        logger.debug('Mode Abstraction:\n' + str(absys) +'\n')

        abstractions[mode] = absys

    # merge their domains
    (merged_abstr, ap_labeling) = merge_partitions(abstractions)
    n = len(merged_abstr.ppp)
    logger.info('Merged partition has: ' + str(n) + ', states')

    # find feasible transitions over merged partition
    trans = dict()
    for mode in modes:
        cont_dyn = hybrid_sys.dynamics[mode]

        params = disc_params[mode]

        trans[mode] = get_transitions(
            merged_abstr, mode, cont_dyn,
            N=params['N'], trans_length=params['trans_length']
        )

    # merge the abstractions, creating a common TS
    merge_abstractions(merged_abstr, trans,
                       abstractions, modes, mode_nums)

    if plot:
        plot_mode_partitions(merged_abstr, show_ts, only_adjacent)

    return merged_abstr

def plot_mode_partitions(swab, show_ts, only_adjacent):
    """Save each mode's partition and final merged partition.
    """
    axs = swab.plot(show_ts, only_adjacent)

    if not axs:
        logger.error('failed to plot the partitions.')
        return

    n = len(swab.modes)
    assert(len(axs) == 2*n)

    # annotate
    for ax in axs:
        plot_annot(ax)

    # save mode partitions
    for ax, mode in zip(axs[:n], swab.modes):
        fname = 'merged_' + str(mode) + '.pdf'
        ax.figure.savefig(fname)

    # save merged partition
    for ax, mode in zip(axs[n:], swab.modes):
        fname = 'part_' + str(mode) + '.pdf'
        ax.figure.savefig(fname)

def plot_annot(ax):
    fontsize = 5

    for tick in ax.xaxis.get_major_ticks():
        tick.label1.set_fontsize(fontsize)
    for tick in ax.yaxis.get_major_ticks():
        tick.label1.set_fontsize(fontsize)
    ax.set_xlabel('$v_1$', fontsize=fontsize+6)
    ax.set_ylabel('$v_2$', fontsize=fontsize+6)

def merge_abstractions(merged_abstr, trans, abstr, modes, mode_nums):
    """Construct merged transitions.

    @type merged_abstr: L{AbstractSwitched}
    @type abstr: dict of L{AbstractPwa}
    """
    # TODO: check equality of atomic proposition sets
    aps = abstr[modes[0]].ts.atomic_propositions

    logger.info('APs: ' + str(aps))

    sys_ts = trs.FTS()

    # create stats
    n = len(merged_abstr.ppp)
    states = range(n)
    sys_ts.states.add_from(states)

    sys_ts.atomic_propositions.add_from(aps)

    # copy AP labels from regions to discrete states
    ppp2ts = states
    for (i, state) in enumerate(ppp2ts):
        props =  merged_abstr.ppp[i].props
        sys_ts.states[state]['ap'] = props

    # create mode actions
    sys_actions = [str(s) for e,s in modes]
    env_actions = [str(e) for e,s in modes]

    # no env actions ?
    if mode_nums[0] == 0:
        actions_per_mode = {
            (e,s):{'sys_actions':str(s)}
            for e,s in modes
        }
        sys_ts.sys_actions.add_from(sys_actions)
    elif mode_nums[1] == 0:
        # no sys actions
        actions_per_mode = {
            (e,s):{'env_actions':str(e)}
            for e,s in modes
        }
        sys_ts.env_actions.add_from(env_actions)
    else:
        actions_per_mode = {
            (e,s):{'env_actions':str(e), 'sys_actions':str(s)}
            for e,s in modes
        }
        sys_ts.env_actions.add_from([str(e) for e,s in modes])
        sys_ts.sys_actions.add_from([str(s) for e,s in modes])

    for mode in modes:
        env_sys_actions = actions_per_mode[mode]
        adj = trans[mode]

        sys_ts.transitions.add_adj(
            adj = adj,
            adj2states = states,
            **env_sys_actions
        )

    merged_abstr.ts = sys_ts
    merged_abstr.ppp2ts = ppp2ts

def get_transitions(
    abstract_sys, mode, ssys, N=10,
    closed_loop=True,
    trans_length=1
):
    """Find which transitions are feasible in given mode.

    Used for the candidate transitions of the merged partition.

    @rtype: scipy.sparse.lil_matrix
    """
    logger.info('checking which transitions remain feasible after merging')
    part = abstract_sys.ppp

    # Initialize matrix for pairs to check
    IJ = part.adj.copy()
    if trans_length > 1:
        k = 1
        while k < trans_length:
            IJ = np.dot(IJ, part.adj)
            k += 1
        IJ = (IJ > 0).astype(int)

    # Initialize output
    n = len(part)
    transitions = sp.lil_matrix((n, n), dtype=int)

    # Do the abstraction
    n_checked = 0
    n_found = 0
    while np.sum(IJ) > 0:
        n_checked += 1

        ind = np.nonzero(IJ)
        i = ind[1][0]
        j = ind[0][0]
        IJ[j,i] = 0

        logger.debug('checking transition: ' + str(i) + ' -> ' + str(j))

        si = part[i]
        sj = part[j]

        # Use original cell as trans_set
        trans_set = abstract_sys.ppp2pwa(mode, i)[1]
        active_subsystem = abstract_sys.ppp2sys(mode, i)[1]

        trans_feasible = is_feasible(
            si, sj, active_subsystem, N,
            closed_loop = closed_loop,
            trans_set = trans_set
        )

        if trans_feasible:
            transitions[i, j] = 1
            msg = '\t Feasible transition.'
            n_found += 1
        else:
            transitions[i, j] = 0
            msg = '\t Not feasible transition.'
        logger.debug(msg)
    logger.info('Checked: ' + str(n_checked))
    logger.info('Found: ' + str(n_found))
    logger.info('Survived merging: ' + str(float(n_found) / n_checked) + ' % ')

    return transitions

def multiproc_merge_partitions(abstractions):
    """LOGTIME in #processors parallel merging.

    Assuming sufficient number of processors.

    UNDER DEVELOPMENT; function signature may change without notice.
    Calling will result in NotImplementedError.
    """
    raise NotImplementedError

def merge_partitions(abstractions):
    """Merge multiple abstractions.

    @param abstractions: keyed by mode
    @type abstractions: dict of L{AbstractPwa}

    @return: (merged_abstraction, ap_labeling)
        where:
            - merged_abstraction: L{AbstractSwitched}
            - ap_labeling: dict
    """
    if len(abstractions) == 0:
        warnings.warn('Abstractions empty, nothing to merge.')
        return

    # consistency check
    for ab1 in abstractions.itervalues():
        for ab2 in abstractions.itervalues():
            p1 = ab1.ppp
            p2 = ab2.ppp

            if p1.prop_regions != p2.prop_regions:
                msg = 'merge: partitions have different sets '
                msg += 'of continuous propositions'
                raise Exception(msg)

            if not (p1.domain.A == p2.domain.A).all() or \
            not (p1.domain.b == p2.domain.b).all():
                raise Exception('merge: partitions have different domains')

            # check equality of original PPP partitions
            if ab1.orig_ppp == ab2.orig_ppp:
                logger.info('original partitions happen to be equal')

    init_mode = abstractions.keys()[0]
    all_modes = set(abstractions)
    remaining_modes = all_modes.difference(set([init_mode]))

    print('init mode: ' + str(init_mode))
    print('all modes: ' + str(all_modes))
    print('remaining modes: ' + str(remaining_modes))

    # initialize iteration data
    prev_modes = [init_mode]

   	# Create a list of merged-together regions
    ab0 = abstractions[init_mode]
    regions = list(ab0.ppp)
    parents = {init_mode:range(len(regions) )}
    ap_labeling = {i:reg.props for i,reg in enumerate(regions)}
    for cur_mode in remaining_modes:
        ab2 = abstractions[cur_mode]
        r = merge_partition_pair(
            regions, ab2, cur_mode, prev_modes,
            parents, ap_labeling
        )
        regions, parents, ap_labeling = r
        prev_modes += [cur_mode]
    new_list = regions

    # build adjacency based on spatial adjacencies of
    # component abstractions.
    # which justifies the assumed symmetry of part1.adj, part2.adj
	# Basically, if two regions are either 1) part of the same region in one of
	# the abstractions or 2) adjacent in one of the abstractions, then the two
	# regions are adjacent in the switched dynamics.
    n_reg = len(new_list)

    adj = np.zeros([n_reg, n_reg], dtype=int)
    for i, reg_i in enumerate(new_list):
        for j, reg_j in enumerate(new_list[0:i]):
            touching = False
            for mode in abstractions:
                pi = parents[mode][i]
                pj = parents[mode][j]

                part = abstractions[mode].ppp

                if (part.adj[pi, pj] == 1) or (pi == pj):
                    touching = True
                    break

            if not touching:
                continue

            if pc.is_adjacent(reg_i, reg_j):
                adj[i,j] = 1
                adj[j,i] = 1
        adj[i,i] = 1

    ppp = PropPreservingPartition(
        domain=ab0.ppp.domain,
        regions=new_list,
        prop_regions=ab0.ppp.prop_regions,
        adj=adj
    )

    abstraction = AbstractSwitched(
        ppp=ppp,
        modes=abstractions,
        ppp2modes=parents,
    )

    return (abstraction, ap_labeling)

def merge_partition_pair(
    old_regions, ab2,
    cur_mode, prev_modes,
    old_parents, old_ap_labeling
):
    """Merge an Abstraction with the current partition iterate.

    @param old_regions: A list of C{Region} that is from either:
        1. The ppp of the first (initial) L{AbstractPwa} to be merged.
        2. A list of already-merged regions
    @type old_regions: list of C{Region}

    @param ab2: Abstracted piecewise affine dynamics to be merged into the
    @type ab2: L{AbstractPwa}

    @param cur_mode: mode to be merged
    @type cur_mode: tuple

    @param prev_modes: list of modes that have already been merged together
    @type prev_modes: list of tuple

    @param old_parents: dict of modes that have already been merged to dict of
        indices of new regions to indices of regions
    @type old_parents: dict of modes to list of region indices in list
        C{old_regions} or dict of region indices to regions in original ppp for
        that mode

    @param old_ap_labeling: dict of states of already-merged modes to sets of
        propositions for each state
    @type old_ap_labeling: dict of tuples to sets

    @return: the following:
        - C{new_list}, list of new regions
        - C{parents}, same as input param C{old_parents}, except that it
          includes the mode that was just merged and for list of regions in
          return value C{new_list}
        - C{ap_labeling}, same as input param C{old_ap_labeling}, except that it
          includes the mode that was just merged.
    """
    logger.info('merging partitions')

    part2 = ab2.ppp

    modes = prev_modes + [cur_mode]

    new_list = []
    parents = {mode:dict() for mode in modes}
    ap_labeling = dict()

    for i in xrange(len(old_regions)):
        for j in xrange(len(part2)):
            isect = pc.intersect(old_regions[i],
                                 part2[j])
            rc, xc = pc.cheby_ball(isect)

            # no intersection ?
            if rc < 1e-5:
                continue
            logger.info('merging region: A' + str(i) +
                        ', with: B' + str(j))

            # if Polytope, make it Region
            if len(isect) == 0:
                isect = pc.Region([isect])

            # label the Region with propositions
            isect.props = old_regions[i].props.copy()

            new_list.append(isect)
            idx = new_list.index(isect)

            # keep track of parents
            for mode in prev_modes:
                parents[mode][idx] = old_parents[mode][i]
            parents[cur_mode][idx] = j

            # union of AP labels from parent states
            ap_label_1 = old_ap_labeling[i]
            ap_label_2 = ab2.ts.states[j]['ap']

            logger.debug('AP label 1: ' + str(ap_label_1))
            logger.debug('AP label 2: ' + str(ap_label_2))

            # original partitions may be different if pwa_partition used
            # but must originate from same initial partition,
            # i.e., have same continuous propositions, checked above
            #
            # so no two intersecting regions can have different AP labels,
            # checked here
            if ap_label_1 != ap_label_2:
                msg = 'Inconsistent AP labels between intersecting regions\n'
                msg += 'of partitions of switched system.'
                raise Exception(msg)

            ap_labeling[idx] = ap_label_1

    return new_list, parents, ap_labeling



#!/usr/bin/env python
# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Convert from most of SPIN syntax to LTL2DSTAR syntax.

Example usage:

  $ ./spin2dstar.py '([]<>a) -> ([]<>b)'


SCL; 3 Sep 2013
"""

import sys
from pyparsing import * #Literal, Forward, Word, alphanums, alphas


SPIN_SYN = {"&": "&&",
            "|": "||",
            "!": "!",
            "->": "->",
            "<->": "<->",
            "[]": "[]",
            "<>": "<>",
            "X": "X",
            "U": "U"}

DSTAR_SYN = {"&": "&",
             "|": "|",
             "!": "!",
             "->": "i",
             "<->": "e",
             "[]": "G",
             "<>": "F",
             "X": "X",
             "U": "U"}


class AST(object):
    def __init__(self, sym="", left="", right=""):
        self.sym = sym
        self.left = left
        self.right = right

    def __str__(self):
        return "("+str(self.left)+str(self.sym)+str(self.right)+")"

    def infix_to_tree(self, toks):
        print(toks)

    def to_dstar(self, pretty=False):
        try:
            output = DSTAR_SYN[self.sym]
        except KeyError:
            output = self.sym
        if self.left != "":
            if pretty:
                output += " "
            output += self.left.to_dstar(pretty=pretty)
        if self.right != "":
            if pretty:
                output += " "
            output += self.right.to_dstar(pretty=pretty)
        return output


def reverse_lookup(d, value):
    """Return first key found with given value.

    Raise ValueError exception if no matches found.
    """
    for (k,v) in d.iteritems():
        if v == value:
            return k

expr_stack = []
def push_op(toks):
    canonized_tok = reverse_lookup(SPIN_SYN, toks[0])
    if canonized_tok in ("!", "[]", "<>", "X"):  # Unary
        if len(expr_stack) < 1:
            raise ValueError("Insufficient number of operands in expression stack")
        expr_stack.append(AST(sym=canonized_tok, right=expr_stack.pop()))
    else:  # Binary
        if len(expr_stack) < 2:
            raise ValueError("Insufficient number of operands in expression stack")
        expr_stack.append(AST(sym=canonized_tok, right=expr_stack.pop(), left=expr_stack.pop()))

def push_ident(toks):
    expr_stack.append(AST(toks[0]))


def spin_to_AST(inform):
    spin_G = Literal(SPIN_SYN["[]"])
    spin_F = Literal(SPIN_SYN["<>"])
    spin_X = Literal(SPIN_SYN["X"])
    spin_U = Literal(SPIN_SYN["U"])
    spin_negate = Literal(SPIN_SYN["!"])
    spin_and = Literal(SPIN_SYN["&"])
    spin_or = Literal(SPIN_SYN["|"])
    spin_implies = Literal(SPIN_SYN["->"])
    spin_equiv = Literal(SPIN_SYN["<->"])

    identifier = Word(alphas+"_", alphanums+"_")

    form = Forward()
    unary_or_less = Forward()
    unary_or_less << ((spin_negate + unary_or_less).setParseAction(push_op)
                      | (spin_X + unary_or_less).setParseAction(push_op)
                      | (spin_G + unary_or_less).setParseAction(push_op)
                      | (spin_F + unary_or_less).setParseAction(push_op)
                      | identifier.setParseAction(push_ident)
                      | (Suppress("(") + form + Suppress(")")))
    form << ((unary_or_less + ZeroOrMore(((spin_U | spin_and | spin_or | spin_implies | spin_equiv) + form).setParseAction(push_op)))
             | unary_or_less)

    form.parseString(inform, parseAll=True)

    assert len(expr_stack) == 1
    return expr_stack.pop()

def spin_to_dstar(inform):
    expr = spin_to_AST(inform)
    return expr.to_dstar(pretty=True)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: "+str(sys.argv[0])+" FORMULA")
        exit(1)
    print(spin_to_dstar(sys.argv[1]) )



#!/usr/bin/env python
from __future__ import print_function
from tulip.spec.form import LTL
from tulip.interfaces import lily

f = LTL('([]<>(!a)) -> ([](a -> <> b) && [](a -> !X b))',
        input_variables={'a': 'boolean'},
        output_variables={'b': 'boolean'})
# Try appending
#
#     && [](!b -> !X b)
#
# to the above LTL formula. The result will not be realizable.

M = lily.synthesize(f)
if M is not None:
    print(M)
else:
    print('Not realizable.')



import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('tulip.ltl_parser_log').setLevel(logging.ERROR)
# import nose.tools as nt
from tulip.spec.ast import nodes
from tulip.spec.parser import parse
from tulip.spec import transformation as tx


def test_to_labeled_graph():
    f = ('( ( p & q ) U ( ( q | ( ( p -> w ) & ( ! ( z -> b ) ) ) ) & '
         '( G ( X g ) ) ) )')
    tree = parse(f)
    assert len(tree) == 18
    nodes = {'p', 'q', 'w', 'z', 'b', 'g', 'G',
             'U', 'X', '&', '|', '!', '->'}
    g = tx.Tree.from_recursive_ast(tree)
    h = tx.ast_to_labeled_graph(g, detailed=False)
    labels = {d['label'] for u, d in h.nodes_iter(data=True)}
    assert labels == nodes


def tree_from_recursive_ast_test():
    # (1 - x) + y
    x = nodes.Var('x')
    one = nodes.Num('1')
    minus = nodes.Operator('-', one, x)
    y = nodes.Var('y')
    plus = nodes.Operator('+', minus, y)
    # convert to Tree
    g = tx.Tree.from_recursive_ast(plus)
    assert len(g) == 5
    assert set(g.nodes()) == set([x, one, minus, y, plus])
    assert g.has_edge(plus, y)
    assert g.has_edge(plus, minus)
    assert g.has_edge(minus, one)
    assert g.has_edge(minus, x)


def tree_to_recursive_ast_test():
    g = tx.Tree()
    x = nodes.Var('x')
    one = nodes.Num('1')
    minus = nodes.Operator('-', one, x)
    y = nodes.Var('y')
    # (1 - x) + y
    plus = nodes.Operator('+', minus, y)
    g.add_nodes_from([plus, x, y, one, minus])
    # g is different than forula:
    # x - (z + 1)
    z = nodes.Var('z')
    g.root = minus
    g.add_edge(minus, x, key=0)
    g.add_edge(minus, plus, key=1)
    g.add_edge(plus, z, key=0)
    g.add_edge(plus, one, key=1)
    t = g.to_recursive_ast()
    # -
    assert isinstance(t, nodes.Operator)
    assert t.operator == '-'
    # must be a new object
    assert t is not minus
    assert len(t.operands) == 2
    # x
    u = t.operands[0]
    assert isinstance(u, nodes.Var)
    assert u.value == 'x'
    assert u is not x
    # +
    u = t.operands[1]
    assert isinstance(u, nodes.Operator)
    assert u.operator == '+'
    assert u is not plus
    assert len(u.operands) == 2
    # z
    v = u.operands[0]
    assert isinstance(v, nodes.Var)
    assert v.value == 'z'
    assert v is not z
    # 1
    u = u.operands[1]
    assert isinstance(u, nodes.Num)
    assert u.value == '1'
    assert u is not one


def test_str_to_int():
    r = parse('a = "hehe"')
    g = tx.Tree.from_recursive_ast(r)
    var_str2int = {'a': ['hehe', 'haha']}
    tx.sub_constants(g, var_str2int)
    f = g.to_recursive_ast()
    s = f.flatten()
    assert s == '( a = 0 )'

    x = '(loc = "s2") -> X((((env_alice = "left") && (env_bob = "bright"))))'
    var_str2int = {
        'loc': ['s0', 's2'],
        'env_alice': ['left', 'right'],
        'env_bob': ['bleft', 'bright']}
    r = parse(x)
    print(repr(r))
    g = tx.Tree.from_recursive_ast(r)
    print(g)
    tx.sub_constants(g, var_str2int)
    print(str(g))
    f = g.to_recursive_ast()
    print(repr(f))
    s = f.flatten()
    print(s)
    assert s == ('( ( loc = 1 ) -> '
                 '( X ( ( env_alice = 0 ) & ( env_bob = 1 ) ) ) )')



#!/usr/bin/env python
"""
Tests for transys.automata (part of transys subpackage)
"""

from tulip import transys as trs


def rabin_test():
    dra = trs.DRA()
    print(dra)

    dra.states.add_from(range(10) )

    dra.states.accepting.add([1,2], [3,4] )
    assert(isinstance(dra.states.accepting._pairs, list) )
    assert(dra.states.accepting._pairs[0][0]._set == {1,2} )
    assert(dra.states.accepting._pairs[0][0]._list == [] )
    assert(dra.states.accepting._pairs[0][1]._set == {3,4} )
    assert(dra.states.accepting._pairs[0][1]._list == [] )
    print(dra.states.accepting)

    dra.states.accepting.remove([1,2], [3,4] )

    assert(isinstance(dra.states.accepting._pairs, list) )
    assert(not dra.states.accepting._pairs)

    dra.states.accepting.add([1], [2, 4] )
    dra.states.accepting.add_states(0, [2], [] )

    dra.states.accepting.add([2, 1], [] )

    assert(isinstance(dra.states.accepting._pairs, list) )
    assert(dra.states.accepting._pairs[0][0]._set == {1,2} )
    assert(dra.states.accepting._pairs[0][0]._list == [] )
    assert(dra.states.accepting._pairs[0][1]._set == {2,4} )
    assert(dra.states.accepting._pairs[0][1]._list == [] )

    assert(dra.states.accepting._pairs[1][0]._set == {2,1} )
    assert(dra.states.accepting._pairs[1][0]._list == [] )
    assert(dra.states.accepting._pairs[1][1]._set == set() )
    assert(dra.states.accepting._pairs[1][1]._list == [] )



#!/usr/bin/env python
"""
Tests for transys.machines (part of transys subpackage)
"""
import logging
logging.basicConfig()
logger = logging.getLogger(__name__)

from tulip.transys import machines

def test_strip_ports():
    mealy = machines.MealyMachine()
    mealy.add_inputs({'door':{'open', 'closed'}})
    mealy.add_outputs({'led':{'on', 'off'},
                       'window':{'open', 'closed'}})

    mealy.add_nodes_from(xrange(10))
    mealy.states.initial.add(0)

    mealy.add_edge(0, 1, door='open', window='open', led='on')
    mealy.add_edge(0, 2, door='open', window='open', led='off')
    mealy.add_edge(2, 3, door='closed', window='closed', led='on')
    mealy.add_edge(3, 4, door='open', window='closed', led='on')
    mealy.add_edge(9, 5, door='closed', window='open', led='off')

    new = machines.strip_ports(mealy, {'window'})

    assert('door' in new.inputs)
    assert('led' in new.outputs)
    assert('window' not in new.outputs)

    edges = [
        (0, 1, dict(door='open', led='on')),
        (0, 2, dict(door='open', led='off')),
        (2, 3, dict(door='closed', led='on')),
        (3, 4, dict(door='open', led='on')),
        (9, 5, dict(door='closed', led='off')),
    ]

    assert(len(edges) == len(new.edges()))
    for (u, v, d), (x, y, b) in zip(new.edges_iter(data=True), edges):
        assert(u == x)
        assert(v == y)
        assert(d == b)



"""
McNaughton's algorithm for solving parity games

Reference
=========
Robert McNaughton (1993). Infinite games played on finite graphs.
Annals of Pure and Applied Logic 65(2): 149--184.
doi:10.1016/0168-0072(93)90036-D
"""
import copy
from tulip import transys as trs

def McNaughton(p):
    """Solve parity game.

    @type p: L{ParityGame}
    """
    p = copy.deepcopy(p)

    if p.has_deadends():
        raise Exception('The GameGraph has sinks !')

    c = p.max_color
    print('highest color occurring in P: c = ' + str(c))

    if c is 0 or len(p) == 0:
        return (set(p), set())

    sigma = c % 2
    W = {0:set(), 1:set()}
    while True:
        max_color_pos = {x for x in p if p.node[x]['color'] == c}

        a = attractor(max_color_pos, p, sigma)

        p_ = copy.deepcopy(p)
        p_.states.remove_from(a)

        W_ = McNaughton(p_)

        if not W_[1-sigma]:
            W[sigma] = set(p) - W[1-sigma]
            return (W[0], W[1])

        a = attractor(W_[1-sigma], p, 1-sigma)
        W[1-sigma] = W[1-sigma].union(a)

        p.states.remove_from(a)

def attractor(W, pg, player):
    """Find attractor set.

    @param w: winning region
    @type w: set

    @type pg: L{ParityGame}

    @param player: for whom to calculate attractor
    """
    W = set(W)
    Wold = set()

    # least fixed point
    while W != Wold:
        Wold = W
        W = W.union(ctrl_next(W, pg, player))
    return W

def ctrl_next(W, pg, player):
    """Find controlled predecessor set.

    These are the nodes that in one step:

      - can reach W because player controls them
      - will reach W because opponent controls them,
        but has no other choice than next(W)
    """
    cnext = set()
    for node in W:
        for pred in pg.predecessors_iter(node):
            if pg.node[pred]['player'] == player:
                print('controlled by player, good')
            elif len(pg.succ[pred]) == 1:
                print('controlled by opponent, bad only 1 outgoing')
            else:
                print('not in CNext')
                continue

            cnext.add(pred)
    return cnext

if __name__ == '__main__':
    p = trs.automata.ParityGame(c=3)

    p.states.add('p0', player=0, color=1)
    p.states.add('p1', player=1, color=0)
    p.states.add('p2', player=1, color=1)

    p.transitions.add('p0', 'p1')
    p.transitions.add('p1', 'p0')
    p.transitions.add('p0', 'p2')
    p.transitions.add('p2', 'p2')

    print(p)
    W0, W1 = McNaughton(p)

    print('player 0 wins from: W0 = ' + str(W0))
    print('player 1 wins from: W1 = ' + str(W1))



#!/usr/bin/env python
"""
Regression tests for spin2dstar.py

SCL; 1 Sep 2013
"""

import sys
import nose

from spin2dstar import spin_to_dstar


S2D_BATTERY = {
    "[]<> a": "G F a",
    "foo": "foo",
    "p0 && p1": "& p0 p1",
    "([] <> a) -> ([] <> b)": "i G F a G F b",
    "(p U q) || (r U z)": "| U p q U r z"
}


def check_spin_to_dstar(informula, expected_outformula):
    assert expected_outformula == spin_to_dstar(informula)

def test_spin_to_dstar():
    for (k,v) in S2D_BATTERY.items():
        yield check_spin_to_dstar, k, v


if __name__ == "__main__":
    sys.argv.append("--verbose")
    nose.run()



# Copyright (c) 2011-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Formulae constituting specifications

what this module adds to the pure syntactic manipulations of
L{transformation} is a variable table, with game semantics in particular.

Also, those modules are for manipulating expressions.
This module knows about program structure,
namely the sections of a specification file.
"""
from __future__ import absolute_import
import logging
logger = logging.getLogger(__name__)
import pprint
import time
import re
import copy
from tulip.spec import parser
from tulip.spec import transformation as tx
from tulip.spec import translation as ts


class LTL(object):
    """LTL formula (specification)

    Minimal class that describes an LTL formula in the canonical TuLiP
    syntax.  It contains three attributes:

      - C{formula}: a C{str} of the formula.  Syntax is only enforced
        if the user requests it, e.g., using the L{check_form} method.

      - C{input_variables}: a C{dict} of variables (names given as
        strings) and their domains; each key is a variable name and
        its value (in the dictionary) is its domain.  See notes below.
        Semantically, these variables are considered to be inputs
        (i.e., uncontrolled, externally determined).

      - C{output_variables}: similar to C{input_variables}, but
        considered to be outputs, i.e., controlled, the strategy for
        setting of which we seek in formal synthesis.

    N.B., domains are specified in multiple datatypes.  The type is
    indicated below in parenthesis.  Recognized domains, along with
    examples, are:

      - boolean (C{str}); this domain is specified by C{"boolean"};
      - finite_set (C{set}); e.g., C{{1,3,5}};
      - range (C{tuple} of length 2); e.g., C{(0,15)}.

    As with the C{formula} attribute, type-checking is only performed
    if requested by the user.  E.g., any iterable can act as a
    finite_set.  However, a range domain must be a C{tuple} of length
    2; otherwise it is ambiguous with finite_set.
    """

    def __init__(self, formula=None, input_variables=None,
                 output_variables=None):
        """Instantiate an LTL object.

        Any non-None arguments are saved to the corresponding
        attribute by reference.
        """
        if formula is None:
            formula = ''
        if input_variables is None:
            input_variables = dict()
        if output_variables is None:
            output_variables = dict()

        self.formula = formula
        self.input_variables = input_variables
        self.output_variables = output_variables

        self.check_vars()

    def __repr__(self):
        return ('LTL(\'{f}\', input_variables={inputv}, ' +
                'output_variables={outv})').format(f=self.formula,
                                                   inputv=self.input_variables,
                                                   outv=self.output_variables)

    def __str__(self):
        return str(self.formula)

    def _domain_str(self, d):
        if d == 'boolean':
            return d
        elif isinstance(d, tuple) and len(d) == 2:
            return '[' + str(d[0]) + ', ' + str(d[1]) + ']'
        elif hasattr(d, '__iter__'):
            return '{' + ', '.join([str(e) for e in d]) + '}'
        else:
            raise ValueError('Unrecognized variable domain type.')

    def dumps(self, timestamp=False):
        """Dump TuLiP LTL file string.

        @param timestamp: If True, then add comment to file with
            current time in UTC.
        """
        if timestamp:
            output = '# Generated at ' + time.strftime(
                '%Y-%m-%d %H:%M:%S UTC', time.gmtime()) + '\n'
        else:
            output = ''
        output += '0  # Version\n%%\n'
        if self.input_variables:
            output += 'INPUT:\n'
            for k, v in self.input_variables.items():
                output += str(k) + ' : ' + self._domain_str(v) + ';\n'
        if self.output_variables:
            output += '\nOUTPUT:\n'
            for k, v in self.output_variables.items():
                output += str(k) + ' : ' + self._domain_str(v) + ';\n'
        return output + '\n%%\n' + self.formula

    def check_vars(self):
        """Raise Exception if variabe definitions are invalid.

        Checks:

          - env and sys have common vars
          - some var is also a possible value of some var (including itself)
            (of arbitrary finite data type)
        """
        common_vars = {x for x in self.input_variables
                       if x in self.output_variables}
        if common_vars:
            raise Exception('Env and sys have variables in common: ' +
                            str(common_vars))

    def check_form(self, check_undeclared_identifiers=False):
        """Verify formula syntax and type-check variable domains.

        Return True iff OK.

        UNDER DEVELOPMENT; function signature may change without
        notice.  Calling will result in NotImplementedError.
        """
        raise NotImplementedError

    @staticmethod
    def loads(s):
        """Create LTL object from TuLiP LTL file string."""
        for line in s.splitlines():
            line = line.strip()  # Ignore leading and trailing whitespace
            comment_ind = line.find('#')
            if comment_ind > -1:
                line = line[:comment_ind]
            if line:  # Version number is the first nonblank line
                try:
                    version = int(line)
                except ValueError:
                    raise ValueError("Malformed version number.")
                if version != 0:
                    raise ValueError("Unrecognized version number: " +
                                     str(version))
                break
        try:
            s = re.sub(r'#.*(\n|$)', '', s)  # Strip comments
            preamble, declar, formula = s.split('%%\n')
            input_ind = declar.find('INPUT:')
            output_ind = declar.find('OUTPUT:')
            if output_ind == -1:
                output_ind = 0  # Default is OUTPUT
            if input_ind == -1:
                input_section = ''
                output_section = declar[output_ind:]
            elif input_ind < output_ind:
                input_section = declar[input_ind:output_ind]
                output_section = declar[output_ind:]
            else:
                output_section = declar[output_ind:input_ind]
                input_section = declar[input_ind:]
            input_section = input_section.replace('INPUT:', '')
            output_section = output_section.replace('OUTPUT:', '')

            variables = [dict(), dict()]
            sections = [input_section, output_section]
            for i in range(2):
                for var_declar in sections[i].split(';'):
                    if not var_declar.strip():
                        continue
                    name, domain = var_declar.split(':')
                    name = name.strip()
                    domain = domain.lstrip().rstrip(';')
                    if domain[0] == '[':  # Range-type domain
                        domain = domain.split(',')
                        variables[i][name] = (
                            int(domain[0][1:]),
                            int(domain[1][:domain[1].index(']')])
                        )
                    elif domain[0] == '{':  # Finite set domain
                        domain.strip()
                        assert domain[-1] == '}'
                        domain = domain[1:-1]  # Remove opening, closing braces
                        variables[i][name] = list()
                        for elem in domain.split(','):
                            try:
                                variables[i][name].append(int(elem))
                            except ValueError:
                                variables[i][name].append(str(elem))
                        variables[i][name] = set(variables[i][name])
                    elif domain == 'boolean':
                        variables[i][name] = domain
                    else:
                        raise TypeError

        except (ValueError, TypeError):
            raise ValueError('Malformed TuLiP LTL specification string.')

        return LTL(
            formula=formula,
            input_variables=variables[0],
            output_variables=variables[1]
        )

    @staticmethod
    def load(f):
        """Wrap L{loads} for reading from files.

        @param f: file or str.  In the latter case, attempt to open a
            file named "f" read-only.
        """
        if isinstance(f, str):
            f = open(f, 'rU')
        return LTL.loads(f.read())


class GRSpec(LTL):
    r"""GR(1) specification.

    The basic form is::

      (env_init & []env_safety & []<>env_prog_1 & []<>env_prog_2 & ...)
        -> (sys_init & []sys_safety & []<>sys_prog_1 & []<>sys_prog_2 & ...)

    A GRSpec object contains the following attributes:

      - C{moore}: select whether a strategy can see primed
        environment variables.

      - C{plus_one}: select causal implication between
        assumptions and guarantees.

      - C{qinit}: select quantification of initial values for variables:

        - C{'\A \A'}: forall env forall sys
          assume C{env_init}
          C{sys_init} must be empty

        - C{'\A \E'}: forall env exist sys (usually not Moore)
          assume C{env_init} and require C{sys_init}

        - C{'\E \A'}: exist sys forall env
          assume C{env_init} and require C{sys_init}

        - C{'\E \E'}: exist env exist sys
          require C{sys_init}
          C{env_init} must be empty

      - C{env_vars}: alias for C{input_variables} of L{LTL},
        concerning variables that are determined by the environment.

      - C{env_init}: a list of string that specifies the assumption
        about the initial state of the environment.

      - C{env_safety}: a list of string that specifies the assumption
        about the evolution of the environment state.

      - C{env_prog}: a list of string that specifies the justice
        assumption on the environment.

      - C{sys_vars}: alias for C{output_variables} of L{LTL},
        concerning variables that are controlled by the system.

      - C{sys_init}: a list of string that specifies the requirement
        on the initial state of the system.

      - C{sys_safety}: a list of string that specifies the safety
        requirement.

      - C{sys_prog}: a list of string that specifies the progress
        requirement.

    An empty list for any formula (e.g., if env_init = []) is marked
    as "True" in the specification. This corresponds to the constant
    Boolean function, which usually means that subformula has no
    effect (is non-restrictive) on the spec.

    Consult L{GRSpec.__init__} concerning arguments at the time of
    instantiation.
    """

    def __init__(self, env_vars=None, sys_vars=None,
                 env_init='', sys_init='',
                 env_safety='', sys_safety='',
                 env_prog='', sys_prog='',
                 moore=True, plus_one=True,
                 qinit=r'\A \A',
                 parser=parser):
        """Instantiate a GRSpec object.

        Instantiating GRSpec without arguments results in an empty
        formula.  The default domain of a variable is "boolean".

        @type env_vars: dict or iterable
        @param env_vars: If env_vars is a dictionary, then its keys
            should be variable names, and values are domains of the
            corresponding variable.  Else, if env_vars is an iterable,
            then assume all environment variables are C{boolean} (or
            "atomic propositions").  Cf. L{GRSpec} for details.

        @type sys_vars: dict or iterable
        @param sys_vars: Mutatis mutandis, env_vars.

        @param env_init, env_safety, env_prog, sys_init, sys_safety, sys_prog:
            A string or iterable of strings.  An empty string is
            converted to an empty list.  A string is placed in a list.
            iterables are converted to lists.  Cf. L{GRSpec}.

        @type moore: bool
        @type plus_one: bool
        @param qinit: see class docstring
        """
        self.parser = parser
        self._ast = dict()
        self._cache = {
            'string': dict(),
            'jtlv': dict(),
            'gr1c': dict(),
            'slugs': dict()
        }
        self._bool_int = dict()
        self._parts = {
            x + y
            for x in {'env_', 'sys_'}
            for y in {'init', 'safety', 'prog'}
        }
        self.moore = moore
        self.plus_one = plus_one
        self.qinit = qinit

        if env_vars is None:
            env_vars = dict()
        elif not isinstance(env_vars, dict):
            env_vars = dict([(v, 'boolean') for v in env_vars])
        if sys_vars is None:
            sys_vars = dict()
        elif not isinstance(sys_vars, dict):
            sys_vars = dict([(v, 'boolean') for v in sys_vars])

        self.env_vars = copy.deepcopy(env_vars)
        self.sys_vars = copy.deepcopy(sys_vars)

        self.env_init = env_init
        self.sys_init = sys_init
        self.env_safety = env_safety
        self.sys_safety = sys_safety
        self.env_prog = env_prog
        self.sys_prog = sys_prog

        for formula_component in self._parts:
            x = getattr(self, formula_component)

            if isinstance(x, str):
                if not x:
                    setattr(self, formula_component, [])
                else:
                    setattr(self, formula_component, [x])
            elif not isinstance(x, list):
                setattr(self, formula_component, list(x))

            setattr(self, formula_component,
                    copy.deepcopy(getattr(self, formula_component)))

        LTL.__init__(self, formula=self.to_canon(),
                     input_variables=self.env_vars,
                     output_variables=self.sys_vars)

    def __repr__(self):
        args = (',\n\n'.join([
                'env_vars={ev}',
                'sys_vars={sv}',
                'env_init={ei}',
                'sys_init={si}',
                'env_safety={es}',
                'sys_safety={ss}',
                'env_prog={ep}',
                'sys_prog={sp}']).format(
                ev=repr(self.env_vars),
                sv=repr(self.sys_vars),
                ei=repr(self.env_init),
                si=repr(self.sys_init),
                es=repr(self.env_safety),
                ss=repr(self.sys_safety),
                ep=repr(self.env_prog),
                sp=repr(self.sys_prog)))
        return '{cls}({args})'.format(cls=type(self).__name__, args=args)

    def __str__(self):
        return self.to_canon()

    def dumps(self, timestamp=False):
        self.formula = self.to_canon()
        return LTL.dumps(self, timestamp=timestamp)

    @staticmethod
    def loads(s):
        """Create GRSpec object from TuLiP LTL file string.

        UNDER DEVELOPMENT; function signature may change without
        notice.  Calling will result in NotImplementedError.
        """
        raise NotImplementedError

    @staticmethod
    def load(f):
        """Wrap L{loads} for reading from files.

        UNDER DEVELOPMENT; function signature may change without
        notice.  Calling will result in NotImplementedError.
        """
        raise NotImplementedError

    def pretty(self):
        """Return pretty printing string."""

        output = 'ENVIRONMENT VARIABLES:\n'
        if self.env_vars:
            for k, v in self.env_vars.iteritems():
                output += '\t' + str(k) + '\t' + str(v) + '\n'
        else:
            output += '\t(none)\n'

        output += '\nSYSTEM VARIABLES:\n'
        if self.sys_vars:
            for k, v in self.sys_vars.iteritems():
                output += '\t' + str(k) + '\t' + str(v) + '\n'
        else:
            output += '\t(none)\n'

        output += '\nFORMULA:\n'

        output += 'ASSUMPTION:\n'
        if self.env_init:
            output += (
                '    INITIAL\n\t  ' +
                '\n\t& '.join([
                    '(' + f + ')' for f in self.env_init
                ]) + '\n'
            )
        if self.env_safety:
            output += (
                '    SAFETY\n\t  []' +
                '\n\t& []'.join([
                    '(' + f + ')' for f in self.env_safety
                ]) + '\n'
            )
        if self.env_prog:
            output += (
                '    LIVENESS\n\t  []<>' +
                '\n\t& []<>'.join([
                    '(' + f + ')' for f in self.env_prog
                ]) + '\n'
            )

        output += 'GUARANTEE:\n'
        if self.sys_init:
            output += (
                '    INITIAL\n\t  ' +
                '\n\t& '.join([
                    '(' + f + ')' for f in self.sys_init
                ]) + '\n'
            )
        if self.sys_safety:
            output += (
                '    SAFETY\n\t  []' +
                '\n\t& []'.join([
                    '(' + f + ')' for f in self.sys_safety
                ]) + '\n'
            )
        if self.sys_prog:
            output += (
                '    LIVENESS\n\t  []<>' +
                '\n\t& []<>'.join([
                    '(' + f + ')' for f in self.sys_prog
                ]) + '\n'
            )
        return output

    def check_syntax(self):
        """Raise `AssertionError` for misplaced primed variables."""
        self._assert_no_primed(self.env_init, 'assumed initial condition')
        self._assert_no_primed(self.sys_init, 'guaranteed initial condition')
        self._assert_no_primed(self.env_prog, 'liveness assumption')
        self._assert_no_primed(self.env_prog, 'liveness guarantee')
        for f in self.env_safety:
            a = self.ast(f)
            primed = tx.collect_primed_vars(a)
            for var in primed:
                if var in self.sys_vars:
                    raise AssertionError(
                        'Syntax error: ' +
                        'primed system variable "{var}"'.format(var=var) +
                        ' found in env safety: {f}'.format(f=f))

    def _assert_no_primed(self, formulae, name):
        """Raise `AssertionError` if primed vars in `formulae`."""
        for f in formulae:
            a = self.ast(f)
            primed = tx.collect_primed_vars(a)
            if primed:
                raise AssertionError(
                    'Syntax error: ' +
                    'primed variables: {primed}'.format(primed=primed) +
                    ' found in {name}: {f}'.format(f=f, name=name))

    def copy(self):
        """Return a copy of `self`."""
        r = GRSpec(
            env_vars=dict(self.env_vars),
            sys_vars=dict(self.sys_vars),
            env_init=copy.copy(self.env_init),
            env_safety=copy.copy(self.env_safety),
            env_prog=copy.copy(self.env_prog),
            sys_init=copy.copy(self.sys_init),
            sys_safety=copy.copy(self.sys_safety),
            sys_prog=copy.copy(self.sys_prog))
        r.moore = self.moore
        r.plus_one = self.plus_one
        r.qinit = self.qinit
        return r

    def __or__(self, other):
        """Create union of two specifications."""
        result = self.copy()

        if not isinstance(other, GRSpec):
            raise TypeError('type(other) must be GRSpec')

        assert self.moore == other.moore, (
            self.moore, other.moore)
        assert self.plus_one == other.plus_one, (
            self.plus_one, other.plus_one)
        assert self.qinit == other.qinit, (
            self.qinit, other.qinit)
        # common vars have same types ?
        for varname in set(other.env_vars) & set(result.env_vars):
            if other.env_vars[varname] != result.env_vars[varname]:
                raise ValueError('Mismatched variable domains')

        for varname in set(other.sys_vars) & set(result.sys_vars):
            if other.sys_vars[varname] != result.sys_vars[varname]:
                raise ValueError('Mismatched variable domains')

        result.env_vars.update(other.env_vars)
        result.sys_vars.update(other.sys_vars)

        for x in self._parts:
            getattr(result, x).extend(getattr(other, x))

        return result

    def to_canon(self):
        """Output formula in TuLiP LTL syntax.

        The format is described in the U{Specifications section
        <http://tulip-control.sourceforge.net/doc/specifications.html>}
        of the TuLiP User's Guide.
        """
        conj_cstr = lambda s: ' && ' if s else ''

        assumption = ''
        if self.env_init:
            assumption += _conj(self.env_init)
        if self.env_safety:
            assumption += conj_cstr(assumption) + _conj(self.env_safety, '[]')
        if self.env_prog:
            assumption += conj_cstr(assumption) + _conj(self.env_prog, '[]<>')

        guarantee = ''
        if self.sys_init:
            guarantee += conj_cstr(guarantee) + _conj(self.sys_init)
        if self.sys_safety:
            guarantee += conj_cstr(guarantee) + _conj(self.sys_safety, '[]')
        if self.sys_prog:
            guarantee += conj_cstr(guarantee) + _conj(self.sys_prog, '[]<>')

        # Put the parts together, simplifying in special cases
        if guarantee:
            if assumption:
                return '(' + assumption + ') -> (' + guarantee + ')'
            else:
                return guarantee
        else:
            return 'True'

    def sub_values(self, var_values):
        """Substitute given values for variables.

        Note that there are three ways to substitute values for variables:

          - syntactic using this function

          - no substitution by user code, instead flatten to python and
            use C{eval} together with a C{dict} defining
            the values of variables, as done in L{eval_init}.

        For converting non-integer finite types to
        integer types, use L{replace_finite_by_int}.

        @return: C{dict} of ASTs after the substitutions,
            keyed by original clause (before substitution).
        """
        logger.info('substitute values for variables...')
        a = copy.deepcopy(self._ast)
        for formula, tree in a.iteritems():
            g = tx.Tree.from_recursive_ast(tree)
            tx.sub_values(g, var_values)
            a[formula] = g.to_recursive_ast()
        logger.info('done with substitutions.\n')
        return a

    def compile_init(self, no_str):
        """Compile python expression for initial conditions.

        The returned bytecode can be used with C{eval}
        and a C{dict} assigning values to variables.
        Its value is the conjunction of C{env_init} and C{sys_init}.

        Use the corresponding python data types
        for the C{dict} values:

              - C{bool} for Boolean variables
              - C{int} for integers
              - C{str} for arbitrary finite types

        @param no_str: if True, then compile the formula
            where all string variables have been replaced by integers.
            Otherwise compile the original formula containing strings.

        @return: python expression compiled for C{eval}
        @rtype: C{code}
        """
        self.str_to_int()
        init = {'env': self.env_init, 'sys': self.sys_init}
        pyinit = dict()
        for side, clauses in init.iteritems():
            if no_str:
                clauses = [self._bool_int[x] for x in clauses]
            logger.info('clauses to compile: ' + str(clauses))
            c = [ts.translate_ast(self.ast(x), 'python').flatten()
                 for x in clauses]
            logger.info('after translation to python: ' + str(c))
            s = _conj(c, op='and')
            if not s:
                s = 'True'
            pyinit[side] = s
        s = 'not ({assumption}) or ({assertion})'.format(
            assumption=pyinit['env'],
            assertion=pyinit['sys'])
        return compile(s, '<string>', 'eval')

    def str_to_int(self):
        """Replace arbitrary finite vars with int vars.

        Returns spec itself if it contains only int vars.
        Otherwise it returns a copy of spec with all arbitrary
        finite vars replaced by int-valued vars.
        """
        logger.info('convert string variables to integers...')
        vars_dict = dict(self.env_vars)
        vars_dict.update(self.sys_vars)
        fvars = {v: d for v, d in vars_dict.iteritems() if isinstance(d, list)}
        # replace symbols by ints
        for p in self._parts:
            for x in getattr(self, p):
                if self._bool_int.get(x) in self._ast:
                    logger.debug(str(x) + ' is in _bool_int cache')
                    continue
                else:
                    logger.debug(str(x) + ' is not in _bool_int cache')
                # get AST
                a = self.ast(x)
                # create AST copy with int and bool vars only
                g = tx.Tree.from_recursive_ast(a)
                tx.sub_constants(g, fvars)
                b = g.to_recursive_ast()
                # formula of int/bool AST
                f = b.flatten()
                self._ast[f] = b  # cache
                # remember map from clauses to int/bool clauses
                self._bool_int[x] = f
        logger.info('done converting to integer variables.\n')

    def ast(self, x):
        """Return AST corresponding to formula x.

        If AST for formula C{x} has already been computed earlier,
        then return cached result.
        """
        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug('current cache of ASTs:\n' +
                         pprint.pformat(self._ast) + 3 * '\n')
            logger.debug('check if: ' + str(x) + ', is in cache.')
        if x in self._ast:
            logger.debug(str(x) + ' is already in cache')
        else:
            logger.info('AST cache does not contain:\n\t' + str(x) +
                        '\nNeed to parse.')
            self.parse()
        return self._ast[x]

    def parse(self):
        """Parse each clause and store it.

        The AST resulting from each clause is stored
        in the C{dict} attribute C{ast}.
        """
        logger.info('parsing ASTs to cache them...')
        vardoms = dict(self.env_vars)
        vardoms.update(self.sys_vars)
        # parse new clauses and cache the resulting ASTs
        for p in self._parts:
            s = getattr(self, p)
            for x in s:
                if x in self._ast:
                    logger.debug(str(x) + ' is already in cache')
                    continue
                logger.debug('parse: ' + str(x))
                tree = self.parser.parse(x)
                g = tx.Tree.from_recursive_ast(tree)
                tx.check_for_undefined_identifiers(g, vardoms)
                self._ast[x] = tree
        # rm cached ASTs that correspond to deleted clauses
        self._collect_cache_garbage(self._ast)
        logger.info('done parsing ASTs.\n')

    def _collect_cache_garbage(self, cache):
        logger.info('collecting garbage from GRSpec cache...')
        # rm cached ASTs that correspond to deleted clauses
        s = set(cache)
        for p in self._parts:
            # emptied earlier ?
            if not s:
                break
            w = getattr(self, p)
            # exclude given formulas
            s.difference_update(w)
            # exclude int/bool-only forms of formulas
            s.difference_update({self._bool_int.get(x) for x in w})
        for x in s:
            cache.pop(x)
        logger.info('cleaned ' + str(len(s)) + ' cached elements.\n')


def replace_dependent_vars(spec, bool2form):
    logger.debug('replacing dependent variables using map:\n\t' +
                 str(bool2form))
    vs = dict(spec.env_vars)
    vs.update(spec.sys_vars)
    logger.debug('variables:\n\t' + str(vs))
    bool2subtree = dict()
    for boolvar, formula in bool2form.iteritems():
        logger.debug('checking var: ' + str(boolvar))
        if boolvar in vs:
            assert vs[boolvar] == 'boolean'
            logger.debug(str(boolvar) + ' is indeed Boolean')
        else:
            logger.debug('spec does not contain var: ' + str(boolvar))
        tree = parser.parse(formula)
        bool2subtree[boolvar] = tx.Tree.from_recursive_ast(tree)
    for s in {'env_init', 'env_safety', 'env_prog',
              'sys_init', 'sys_safety', 'sys_prog'}:
        part = getattr(spec, s)
        new = []
        for clause in part:
            logger.debug('replacing in clause:\n\t' + clause)
            tree = spec.ast(clause)
            g = tx.Tree.from_recursive_ast(tree)
            tx.sub_bool_with_subtree(g, bool2subtree)
            f = g.to_recursive_ast().flatten()
            new.append(f)
            logger.debug('caluse tree after replacement:\n\t' + f)
        setattr(spec, s, new)


def _conj(iterable, unary='', op='&&'):
    return ' {op} '.format(op=op).join(
        ['{u}({s})'.format(u=unary, s=s) for s in iterable])



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Finite State Machines Module"""
from __future__ import absolute_import
import copy
from pprint import pformat
from random import choice
from tulip.transys.labeled_graphs import LabeledDiGraph
# inline imports:
#
# from tulip.transys.export import machine2scxml


_hl = 40 * '-'
# port type
pure = {'present', 'absent'}


def is_valuation(ports, valuations):
    for name, port_type in ports.iteritems():
        curvaluation = valuations[name]
        # functional set membership description ?
        if callable(port_type):
            ok = port_type(curvaluation)
        else:
            ok = curvaluation in port_type
        if not ok:
            raise TypeError('Not a valuation.')


def create_machine_ports(spc_vars):
    """Create proper port domains of valuations, given port types.

    @param spc_vars: port names and types inside tulip.
        For arbitrary finite types the type can be a list of strings,
        instead of a range of integers.
        These are as originally defined by the user or synth.
    """
    ports = dict()
    for env_var, var_type in spc_vars.iteritems():
        if var_type == 'boolean':
            domain = {0, 1}
        elif isinstance(var_type, tuple):
            # integer domain
            start, end = var_type
            domain = set(range(start, end + 1))
        elif isinstance(var_type, list):
            # arbitrary finite domain defined by list var_type
            domain = set(var_type)
        ports[env_var] = domain
    return ports


class Transducer(LabeledDiGraph):
    """Sequential Transducer, i.e., a letter-to-letter function.

    Inputs
    ======
    P = {p1, p2,...} is the set of input ports.
    An input port p takes values in a set Vp.
    Set Vp is called the "type" of input port p.
    A "valuation" is an assignment of values to the input ports in P.

    We call "inputs" the set of pairs::

      {(p_i, Vp_i),...}

    of input ports p_i and their corresponding types Vp_i.

    A guard is a predicate (bool-valued) used as sub-label for a transition.
    A guard is defined by a set and evaluated using set membership.
    So given an input port value p=x, then if::

      x \in guard_set

    then the guard is True, otherwise it is False.

    The "inputs" are defined by an OrderedDict::

      {'p1':explicit, 'p2':check, 'p3':None, ...}

    where:
      - C{explicit}:
        is an iterable representation of Vp,
        possible only for discrete Vp.
        If 'p1' is explicitly typed, then guards are evaluated directly::

          input_port_value == guard_value ?

      - C{check}:
        is a class with methods:

          - C{__contains__(x) }:
            check if guard value given to input port 'p1' is
            in the set of possible values Vp.

          - C{__call__(guard_set, input_port_value) }:
            check if C{input_port_value} \\in C{guard_set}
            This allows symbolic type definitions.

            For example, C{input_port_value} might be assigned
            int values, but the C{guard_set} be defined by
            a symbolic expression as the str: 'x<=5'.

            Then the user is responsible for providing
            the appropriate method to the Mealy Machine,
            using the custom C{check} class described here.

            Note that we could provide a rudimentary library
            for the basic types of checks, e.g., for
            the above simple symbolic case, where using
            function eval() is sufficient.

      - C{None}:
        signifies that no type is currently defined for
        this input port, so input type checking and guard
        evaluation are disabled.

        This can be used to skip type definitions when
        they are not needed by the user.

        However, since Machines are in general the output
        of synthesis, it follows that they are constructed
        by code, so the benefits of typedefs will be
        considerable compared to the required coding effort.

    Guards annotate transitions::

      Guards: States x States ---> Input_Predicates

    Outputs
    =======
    Similarly defined to inputs, but:

      - for Mealy Machines they annotate transitions
      - for Moore Machines they annotate states

    State Variables
    ===============
    Similarly defined to inputs, they annotate states,
    for both Mealy and Moore machines::

      States ---> State_Variables

    Update Function
    ===============
    The transition relation:

      - for Mealy Machines::

        States x Input_Valuations ---> Output_Valuations x States

        Note that in the range Output_Valuations are ordered before States
        to emphasize that an output_valuation is produced
        during the transition, NOT at the next state.

        The data structure representation of the update function is
        by storage of the Guards function and definition of Guard
        evaluation for each input port via the OrderedDict discussed above.

      - for Moore Machines::

        States x Input_Valuations ---> States
        States ---> Output_valuations

    Note
    ====
    A transducer may operate on either finite or infinite words, i.e.,
    it is not equipped with interpretation semantics on the words,
    so it does not "care" about word length.
    It continues as long as its input is fed with letters.

    For Machines, each state label consists of (possibly multiple) sublabels,
    each of which is either a variable, or, only for Moore machines,
    may be an output.

    See Also
    ========
    FSM, MealyMachine, MooreMachine
    """

    def __init__(self):
        # values will point to values of _*_label_def below
        self.state_vars = dict()
        self.inputs = dict()
        self.outputs = dict()
        # self.set_actions = {}

        # state labeling
        self._state_label_def = dict()
        self._state_dot_label_format = {'type?label': ':',
                                        'separator': r'\\n'}

        # edge labeling
        self._transition_label_def = dict()
        self._transition_dot_label_format = {'type?label': ':',
                                             'separator': r'\\n'}
        self._transition_dot_mask = dict()
        self._state_dot_mask = dict()

        self.default_export_fname = 'fsm'

        LabeledDiGraph.__init__(self)

        self.dot_node_shape = {'normal': 'ellipse'}
        self.default_export_fname = 'fsm'

    def add_inputs(self, new_inputs, masks=None):
        """Create new inputs.

        @param new_inputs: C{dict} of pairs {port_name : port_type}
            where:
                - port_name: str
                - port_type: Iterable | check class
        @type new_inputs: dict

        @param masks: custom mask functions, for each sublabel
            based on its current value
            each such function returns:
                - True, if the sublabel should be shown
                - False, otherwise (to hide it)
        @type masks: C{dict} of functions C{{port_name : mask_function}}
            each C{mask_function} returns bool
        """
        for port_name, port_type in new_inputs.iteritems():
            # append
            self._transition_label_def[port_name] = port_type
            # inform inputs
            self.inputs[port_name] = port_type
            # printing format
            self._transition_dot_label_format[port_name] = str(port_name)
            if masks is None:
                continue
            if port_name in masks:
                mask_func = masks[port_name]
                self._transition_dot_mask[port_name] = mask_func

    def add_state_vars(self, new_state_vars):
        for var_name, var_type in new_state_vars.iteritems():
            # append
            self._state_label_def[var_name] = var_type
            # inform state vars
            self.state_vars[var_name] = self._state_label_def[var_name]
            # printing format
            self._state_dot_label_format[var_name] = str(var_name)


class MooreMachine(Transducer):
    """Moore machine.

    A Moore machine implements the discrete dynamics::
        x[k+1] = f(x[k], u[k] )
        y[k] = g(x[k] )
    where:
      - k: discrete time = sequence index
      - x: state = valuation of state variables
      - X: set of states = S
      - u: inputs = valuation of input ports
      - y: output actions = valuation of output ports
      - f: X-> 2^X, transition function
      - g: X-> Out, output function
    Observe that the output depends only on the state.

    Note
    ====
    valuation: assignment of values to each port

    Reference
    =========
    U{[M56]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#m56>}
    """

    def __init__(self):
        """Instantiate a Moore state machine."""
        Transducer.__init__(self)
        self.dot_node_shape = {'normal': 'ellipse'}
        self.default_export_fname = 'moore'

    def __str__(self):
        """Get informal string representation."""
        s = (
            _hl + '\nMoore Machine: ' + self.name + '\n' + _hl + '\n' +
            'State Variables:\n\t(name : type)\n' +
            _print_ports(self.state_vars) +
            'Input Ports:\n\t(name : type)\n' +
            _print_ports(self.inputs) +
            'Output Ports:\n\t(name : type)\n' +
            _print_ports(self.outputs) +
            'States & State Var Values: (state : outputs : vars)\n')
        for state, label_dict in self.states(data=True):
            s += '\t' + str(state) + ' :\n'
            # split into vars and outputs
            var_values = {k: v for k, v in label_dict.iteritems()
                          if k in self.state_vars}
            output_values = {k: v for k, v in label_dict.iteritems()
                             if k in self.outputs}
            s += (_print_label(var_values) + ' : ' +
                  _print_label(output_values))
        s += (
            'Initial States:\n' +
            pformat(self.states.initial, indent=3) + 2 * '\n')
        s += 'Transitions & Labels: (from --> to : label)\n'
        for from_state, to_state, label_dict in self.transitions(data=True):
            s += (
                '\t' + str(from_state) + ' ---> ' +
                str(to_state) + ' :\n' +
                _print_label(label_dict))
        s += _hl + '\n'
        return s

    def add_outputs(self, new_outputs, masks=None):
        for port_name, port_type in new_outputs.iteritems():
            # append
            self._state_label_def[port_name] = port_type
            # inform state vars
            self.outputs[port_name] = port_type
            # printing format
            self._state_dot_label_format[port_name] = \
                '/' + str(port_name)
            if masks is None:
                continue
            if port_name in masks:
                mask_func = masks[port_name]
                self._state_dot_mask[port_name] = mask_func


class MealyMachine(Transducer):
    """Mealy machine.

    Examples
    ========
    Traffic Light: Fig. 3.14, p.72 U{[LS11]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#ls11>}

    >>> m = MealyMachine()
    >>> pure_signal = {'present', 'absent'}
    >>> m.add_inputs([('tick', pure_signal) ])
    >>> m.add_outputs([('go', pure_signal), ('stop', pure_signal) ])
    >>> m.states.add_from(['red', 'green', 'yellow'])
    >>> m.states.initial.add('red')

    For brevity:

    >>> p = 'present'
    >>> a = 'absent'

    The transitions can equivalently be defined with dict().
    So instead of the previous C{m.transitions.add}, we can use:

    >>> label = {'tick':p, 'go':p, 'stop':a}
    >>> m.transitions.add('red', 'green', **label)
    >>> label = {'tick':p, 'go':a, 'stop':p}
    >>> m.transitions.add('green', 'yellow', **label)
    >>> label = {'tick':p, 'go':a, 'stop':p}
    >>> m.transitions.add('yellow', 'red', **label)

    This avoids any ordering issues, i.e., changing the
    order of the sublabels does not matter:

    >>> label = {'go':p, 'tick':p, 'stop':a}
    >>> m.transitions.add('red', 'green', **label)

    Theory
    ======
    A Mealy machine implements the discrete dynamics::
        x[k+1] = f(x[k], u[k] )
        y[k] = g(x[k], u[k] )
    where:
      - k: discrete time = sequence index
      - x: state = valuation of state variables
      - X: set of states = S
      - u: inputs = valuation of input ports
      - y: output actions = valuation of output ports
      - f: X-> 2^X, transition function
      - g: X-> Out, output function
    Observe that the output is defined when a reaction occurs to an input.

    Note
    ====
    valuation: assignment of values to each port

    Reference
    =========
    U{[M55]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#m55>}
    """

    def __init__(self):
        Transducer.__init__(self)
        # will point to selected values of self._transition_label_def
        self.dot_node_shape = {'normal': 'ellipse'}
        self.default_export_fname = 'mealy'

    def __str__(self):
        """Get informal string representation."""
        s = (
            _hl + '\nMealy Machine: ' + self.name + '\n' + _hl + '\n' +
            'State Variables:\n\t(name : type)\n' +
            _print_ports(self.state_vars))
        s += 'States & State Var Values:\n'
        for state, label_dict in self.states(data=True):
            s += ('\t' + str(state) + ' :\n' +
                  _print_label(label_dict))
        s += (
            'Initial States:\n' +
            pformat(self.states.initial, indent=3) + 2 * '\n' +
            'Input Ports:\n\t(name : type)\n' +
            _print_ports(self.inputs) +
            'Output Ports:\n\t(name : type)\n' +
            _print_ports(self.outputs) +
            'Transitions & Labels: (from --> to : label)\n')
        for from_state, to_state, label_dict in self.transitions(data=True):
            s += (
                '\t' + str(from_state) + ' ---> ' +
                str(to_state) + ' :\n' +
                _print_label(label_dict))
        s += _hl + '\n'
        return s

    def _save(self, path, fileformat):
        """Export options available only for Mealy machines.

        @type fileformat: 'scxml'
        """
        if fileformat != 'scxml':
            return False
        from tulip.transys.export import machine2scxml
        s = machine2scxml.mealy2scxml(self)
        # dump to file
        f = open(path, 'w')
        f.write(s)
        f.close()
        return True

    def add_outputs(self, new_outputs, masks=None):
        """Add new outputs.

        @param new_outputs: dict of pairs {port_name : port_type}
          where:
            - port_name: str
            - port_type: Iterable | check class
        @type new_outputs: dict

        @param masks: custom mask functions, for each sublabel
            based on its current value
            each such function returns:
              - True, if the sublabel should be shown
              - False, otherwise (to hide it)
        @type masks: dict of functions
            keys are port_names (see arg: new_outputs)
            each function returns bool
        """
        for port_name, port_type in new_outputs.iteritems():
            # append
            self._transition_label_def[port_name] = port_type
            # inform state vars
            self.outputs[port_name] = \
                self._transition_label_def[port_name]
            # printing format
            self._transition_dot_label_format[port_name] = \
                '/' + str(port_name)
            if masks is None:
                continue
            if port_name in masks:
                mask_func = masks[port_name]
                self._transition_dot_mask[port_name] = mask_func

    def reaction(self, from_state, inputs, lazy=False):
        """Return next state and output, when reacting to given inputs.

        The machine must be deterministic.
        (for each state and input at most a single transition enabled,
        this notion does not coincide with output-determinism)

        Not exactly a wrapper of L{Transitions.find},
        because it matches only that part of an edge label
        that corresponds to the inputs.

        @param from_state: transition starts from this state.
        @type from_state: element of C{self.states}

        @param inputs: C{dict} assigning a valid value to each input port.
        @type inputs: {'port_name':port_value, ...}

        @param lazy: Lazy evaluation of inputs? If lazy=True, then
            allow an incomplete specification of input if there is
            precisely one enabled transition.
        @type lazy: bool

        @return: output values and next state.
        @rtype: (outputs, next_state)
          where C{outputs}: C{{'port_name':port_value, ...}}
        """
        if lazy:
            restricted_inputs = set(self.inputs).intersection(inputs.keys())
        else:
            restricted_inputs = self.inputs
        # match only inputs (explicit valuations, not symbolic)
        enabled_trans = [
            (i, j, d)
            for i, j, d in self.edges_iter([from_state], data=True)
            if project_dict(d, restricted_inputs) == inputs]

        if len(enabled_trans) == 0:
            some_possibilities = []
            for i, j, d in self.edges_iter([from_state], data=True):
                # The number of possible inputs to suggest here is
                # arbitrary. Consider making it a function parameter.
                if len(some_possibilities) >= 5:
                    break
                possible_inputs = project_dict(d, restricted_inputs)
                if possible_inputs not in some_possibilities:
                    some_possibilities.append(possible_inputs)

        # must be deterministic
        try:
            ((_, next_state, attr_dict), ) = enabled_trans
        except ValueError:
            if len(enabled_trans) == 0:
                if len(some_possibilities) == 0:
                    raise Exception(
                        'state {from_state} is a dead-end. '
                        'There are no possible inputs from '
                        'it.'.format(from_state=from_state))
                else:
                    raise Exception(
                        'not a valid input, '
                        'some possible inputs include: '
                        '{t}'.format(t=some_possibilities))
            else:
                raise Exception(
                    'must be input-deterministic, '
                    'found enabled transitions: '
                    '{t}'.format(t=enabled_trans))
        outputs = project_dict(attr_dict, self.outputs)
        return (next_state, outputs)

    def reactionpart(self, from_state, inputs):
        """Wraps reaction() with lazy=True
        """
        return self.reaction(from_state, inputs, lazy=True)

    def run(self, from_state=None, input_sequences=None):
        """Guided or interactive run.

        @param input_sequences: if C{None}, then call L{interactive_run},
            otherwise call L{guided_run}.

        @return: output of L{guided_run}, otherwise C{None}.
        """
        if input_sequences is None:
            interactive_run(self, from_state=from_state)
        else:
            return guided_run(self, from_state=from_state,
                              input_sequences=input_sequences)


def guided_run(mealy, from_state=None, input_sequences=None):
    """Run deterministic machine reacting to given inputs.

    @param from_state: start simulation

    @param mealy: input-deterministic Mealy machine
    @type mealy: L{MealyMachine}

    @param from_state: start simulation at this state.
        If C{None}, then use the unique initial state C{Sinit}.

    @param input_sequences: one sequence of values for each input port
    @type input_sequences: C{dict} of C{lists}

    @return: sequence of states and sequence of output valuations
    @rtype: (states, output_sequences)
      where:
        - C{states} is a C{list} of states excluding C{from_state}
        - C{output_sequences} is a C{dict} of C{lists}
    """
    seqs = input_sequences  # abbrv
    missing_ports = set(mealy.inputs).difference(seqs)
    if missing_ports:
        raise ValueError('missing input port(s): ' + missing_ports)
    # dict of lists ?
    non_lists = {k: v for k, v in seqs.iteritems() if not isinstance(v, list)}
    if non_lists:
        raise TypeError('Values must be lists, for: ' + str(non_lists))
    # uniform list len ?
    if len(set(len(x) for x in seqs.itervalues())) > 1:
        raise ValueError('All input sequences must be of equal length.')
    # note: initial sys state non-determinism not checked
    # initial sys edge non-determinism checked instead (more restrictive)
    if from_state is None:
        state = next(iter(mealy.states.initial))
    else:
        state = from_state
    n = len(next(seqs.itervalues()))
    states_seq = []
    output_seqs = {k: list() for k in mealy.outputs}
    for i in range(n):
        inputs = {k: v[i] for k, v in seqs.iteritems()}
        state, outputs = mealy.reaction(state, inputs)
        states_seq.append(state)
        for k in output_seqs:
            output_seqs[k].append(outputs[k])
    return (states_seq, output_seqs)


def random_run(mealy, from_state=None, N=10):
    """Return run from given state for N random inputs.

    Inputs selected randomly in a way that does not block the machine
    So they are not arbitrarily random.
    If the machine is a valid synthesis solution,
    then all safe environment inputs can be generated this way.

    Randomly generated inputs may violate liveness assumption on environment.

    @param mealy: input-deterministic Mealy machine
    @type mealy: C{MealyMachine}

    @param N: number of reactions (inputs)
    @type N: int

    @return: same as L{guided_run}
    """
    if from_state is None:
        state = next(iter(mealy.states.initial))
    else:
        state = from_state
    states_seq = []
    output_seqs = {k: list() for k in mealy.outputs}
    for i in xrange(N):
        trans = mealy.transitions.find([state])
        # choose next transition
        selected_trans = choice(list(trans))
        _, new_state, attr_dict = selected_trans
        # extend execution trace
        states_seq.append(new_state)
        # extend output traces
        outputs = project_dict(attr_dict, mealy.outputs)
        for k in output_seqs:
            output_seqs[k].append(outputs[k])
        # updates
        old_state = state
        state = new_state
        # printing
        inputs = project_dict(attr_dict, mealy.inputs)
        print(
            'move from\n\t state: ' + str(old_state) +
            '\n\t with input:' + str(inputs) +
            '\n\t to state: ' + str(new_state) +
            '\n\t reacting by producing output: ' + str(outputs))
    return (states_seq, output_seqs)


def interactive_run(mealy, from_state=None):
    """Run input-deterministic Mealy machine using user input.

    @param mealy: input-deterministic Mealy machine
    @type mealy: L{MealyMachine}
    """
    if from_state is None:
        state = next(iter(mealy.states.initial))
    else:
        state = from_state
    while True:
        print('\n Current state: ' + str(state))
        if _interactive_run_step(mealy, state) is None:
            break


def _interactive_run_step(mealy, state):
    if state is None:
        raise Exception('Current state is None')
    # note: the spaghettiness of previous version was caused
    #   by interactive simulation allowing both output-non-determinism
    #   and implementing spawning (which makes sense only for generators,
    #   *not* for transducers)
    trans = mealy.transitions.find([state])
    if not trans:
        print('Stop: no outgoing transitions.')
        return None
    while True:
        try:
            selected_trans = _select_transition(mealy, trans)
        except:
            print('Selection not recognized. Please try again.')
    if selected_trans is None:
        return None
    (from_, to_state, attr_dict) = selected_trans
    inputs = project_dict(attr_dict, mealy.inputs)
    outputs = project_dict(attr_dict, mealy.outputs)
    print(
        'Moving from state: ' + str(state) +
        ', to state: ' + str(to_state) + '\n' +
        'given inputs: ' + str(inputs) + '\n' +
        'reacting with outputs: ' + str(outputs))
    return True


def _select_transition(mealy, trans):
    msg = 'Found more than 1 outgoing transitions:' + 2 * '\n'
    for i, t in enumerate(trans):
        (from_state, to_state, attr_dict) = t
        inputs = project_dict(attr_dict, mealy.inputs)
        outputs = project_dict(attr_dict, mealy.outputs)
        msg += (
            '\t' + str(i) + ' : ' +
            str(from_state) + ' ---> ' + str(to_state) + '\n' +
            '\t inputs:' + str(inputs) +
            '\t outputs:' + str(outputs) +
            '\n\n')
    msg += (
        '\n' +
        'Select from the available transitions above\n' +
        'by giving its integer,\n' +
        'Press "Enter" to stop the simulation:\n' +
        '\t int = ')
    id_selected = raw_input(msg)
    if not id_selected:
        return None
    return trans[int(id_selected)]


def moore2mealy(moore):
    """Convert Moore machine to equivalent Mealy machine.

    Reference
    =========
    U{[LS11]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#ls11>}

    @type moore: L{MooreMachine}

    @rtype: L{MealyMachine}
    """
    if not isinstance(moore, MooreMachine):
        raise TypeError('moore must be a MooreMachine')
    mealy = MealyMachine()
    # cp inputs
    for port_name, port_type in moore.inputs.iteritems():
        mask_func = moore._transition_dot_mask.get(port_name)
        if mask_func is None:
            masks = None
        else:
            masks = {port_name: mask_func}
        mealy.add_inputs({port_name: port_type}, masks=masks)
    # cp outputs
    for port_name, port_type in moore.outputs.iteritems():
        mask_func = moore._state_dot_mask.get(port_name)
        if mask_func is None:
            masks = None
        else:
            masks = {port_name: mask_func}
        mealy.add_outputs({port_name: port_type}, masks=masks)
    # cp states
    mealy.states.add_from(moore.states())
    mealy.states.initial.add_from(moore.states.initial)
    # cp transitions
    for si in moore:
        output_values = {
            k: v for k, v in moore.states[si].iteritems()
            if k in moore.outputs}
        output_values = copy.deepcopy(output_values)
        for si_, sj, attr_dict in moore.transitions.find(si):
            # note that we don't filter only input ports,
            # so other edge annotation is preserved
            attr_dict = copy.deepcopy(attr_dict)
            attr_dict.update(output_values)
            mealy.transitions.add(si, sj, attr_dict)
    return mealy


def mealy2moore(mealy):
    """Convert Mealy machine to almost equivalent Moore machine.

    A Mealy machine cannot be transformed to an equivalent Moore machine.
    It can be converted to a Moore machine with an arbitrary initial output,
    which outputs the Mealy output at its next reaction.

    Reference
    =========
    U{[LS11]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#ls11>}

    @type mealy: L{MealyMachine}

    @rtype: L{MooreMachine}
    """
    # TODO: check for when Mealy is exactly convertible to Moore
    if not isinstance(mealy, MealyMachine):
        raise TypeError('moore must be a MealyMachine')
    moore = MooreMachine()
    # cp inputs
    for port_name, port_type in mealy.inputs.iteritems():
        mask_func = mealy._transition_dot_mask.get(port_name)
        if mask_func is None:
            masks = None
        else:
            masks = {port_name: mask_func}
        moore.add_inputs({port_name: port_type}, masks=masks)
    # cp outputs
    for port_name, port_type in mealy.outputs.iteritems():
        mask_func = mealy._transition_dot_mask.get(port_name)
        if mask_func is None:
            masks = None
        else:
            masks = {port_name: mask_func}
        moore.add_outputs({port_name: port_type}, masks=masks)
    # initial state with arbitrary label
    out = {k: list(v)[0] for k, v in mealy.outputs.iteritems()}
    s0 = list(mealy.states.initial)[0]
    # create maps between Moore and Mealy states
    moore2mealy_states = dict()  # {qj : si} (function)
    mealy2moore_states = dict()  # {si : {qj, qk, ...} } (relation)
    new_s0 = _create_state_str(
        s0, out, moore, moore2mealy_states,
        mealy2moore_states)
    moore.states.add(new_s0, out)
    moore.states.initial.add(new_s0)
    # cp transitions and create appropriate states
    Q = set()
    S = set()
    Q.add(new_s0)
    S.add(new_s0)
    while Q:
        new_si = Q.pop()
        si = moore2mealy_states[new_si]
        for si_, sj, attr_dict in mealy.transitions.find(si):
            in_values, out_values = _split_io(attr_dict, mealy)
            new_sj = _create_state_str(
                sj, out_values, moore, moore2mealy_states,
                mealy2moore_states)
            moore.transitions.add(new_si, new_sj, in_values)
            if new_sj not in S:
                Q.add(new_sj)
                S.add(new_sj)
    return moore


def _print_ports(port_dict):
    s = ''
    for port_name, port_type in port_dict.iteritems():
        s += '\t' + str(port_name) + ' : '
        s += pformat(port_type) + '\n'
    s += '\n'
    return s


def _print_label(label_dict):
    s = ''
    for name, value in label_dict.iteritems():
        s += '\t\t' + str(name) + ' : ' + str(value) + '\n'
    s += '\n'
    return s


def _create_state_str(mealy_state, output, moore,
                      moore2mealy_states,
                      mealy2moore_states):
    """Used to create Moore states when converting Mealy -> Moore."""
    for s in mealy2moore_states.setdefault(mealy_state, set()):
        # check output values
        if moore.states[s] == output:
            return s
    # create new
    n = len(moore)
    s = 's' + str(n)
    moore.states.add(s, output)
    moore2mealy_states[s] = mealy_state
    mealy2moore_states[mealy_state].add(s)
    return s


def _split_io(attr_dict, machine):
    """Split into inputs and outputs."""
    input_values = {k: v for k, v in attr_dict.iteritems()
                    if k in machine.inputs}
    output_values = {k: v for k, v in attr_dict.iteritems()
                     if k in machine.outputs}
    return input_values, output_values


project_dict = lambda x, y: {k: x[k] for k in x if k in y}
trim_dict = lambda x, y: {k: x[k] for k in x if k not in y}


def strip_ports(mealy, names):
    """Remove ports in C{names}.

    For example, to remove the atomic propositions
    labeling the transition system C{ts} used
    (so they are dependent variables), call it as:

      >>> strip_ports(mealy, ts.atomic_propositions)

    @type mealy: L{MealyMachine}

    @type names: iterable container of C{str}
    """
    new = MealyMachine()

    new.add_inputs(trim_dict(mealy.inputs, names))
    new.add_outputs(trim_dict(mealy.outputs, names))

    new.add_nodes_from(mealy)
    new.states.initial.add_from(mealy.states.initial)

    for u, v, d in mealy.edges_iter(data=True):
        d = trim_dict(d, names)
        new.add_edge(u, v, **d)
    return new



"""
Only supports closed-loop and non-conservative simulation.
"""
from tulip import hybrid, abstract
import polytope
import scipy.io
import numpy

def export(filename, mealy_machine, system_dynamics=None, abstraction=None,
    disc_params=None, R=None, r=None, Q=None, mid_weight=None):

    """Creates two matlab files. One is a script that generates a Simulink model
    that contains a Stateflow Chart of the Mealy Machine, a block that contains
    a get_input function, a block that maps continuous state to discrete state,
    and a block that times simulation of get_input.

    @param filename: string ending in '.mat'
    @param system: L{LtiSysDyn}, L{PwaSysDyn}, or L{HybridSysDyn} to be saved in
        the .mat file.
    @param filename: String containing name of the .mat file to be created.
    @rtype: None
    """

    # Check whether we're discrete or continuous
    if ((system_dynamics is None) and (abstraction is None) and
        (disc_params is None)):
        is_continuous = False
    elif ((system_dynamics is not None) and (abstraction is not None) and
          (disc_params is not None)):
        is_continuous = True
    else:
        raise StandardError('Cannot tell whether system is continuous or ' +
            'discrete. Please specify dynamics and abstraciton and ' +
            'discretization parameters or none at all.')

    output = {}
    output['is_continuous'] = is_continuous

    # Only export dynamics and abstraction and control weights if the system is
    # continuous
    if is_continuous:

        # Export system dynamics, get state and input dimension
        if isinstance(system_dynamics, hybrid.LtiSysDyn):
            dynamics_output = lti_export(system_dynamics)
            dynamics_output['type'] = 'LtiSysDyn'
            state_dimension = numpy.shape(system_dynamics.A)[1]
            input_dimension = numpy.shape(system_dynamics.B)[1]
        elif isinstance(system_dynamics, hybrid.PwaSysDyn):
            dynamics_output = pwa_export(system_dynamics)
            dynamics_output['type'] = 'PwaSysDyn'
            state_dimension = numpy.shape(system_dynamics.list_subsys[0].A)[1]
            input_dimension = numpy.shape(system_dynamics.list_subsys[0].B)[1]
        elif isinstance(system_dynamics, hybrid.SwitchedSysDyn):
            dynamics_output = switched_export(system_dynamics)
            dynamics_output['type'] = 'SwitchedSysDyn'

            # getting state and input dimension by looking at size of the A and
            # B matrices of one of the PWA subsystems
            pwa_systems = system_dynamics.dynamics.values()
            pwa_system = pwa_systems[0]
            state_dimension = numpy.shape(pwa_system.list_subsys[0].A)[1]
            input_dimension = numpy.shape(pwa_system.list_subsys[0].B)[1]
        else:
            raise TypeError(str(type(system_dynamics)) +
                ' is not a supported type of system dynamics.')
        output['system_dynamics'] = dynamics_output

        # Control weights. Set default values if needed.
        if R is None:
            R = numpy.zeros([state_dimension, state_dimension])
        if r is None:
            r = numpy.zeros([1, state_dimension])
        if Q is None:
            Q = numpy.eye(input_dimension)
        if mid_weight is None:
            mid_weight = 3

        control_weights = {}
        control_weights['state_weight'] = R
        control_weights['input_weight'] = Q
        control_weights['linear_weight'] = r
        control_weights['mid_weight'] = mid_weight
        output['control_weights'] = control_weights

        # Simulation parameters; insert default discretization values if needed
        sim_params = {}
        try:
            sim_params['horizon'] = disc_params['N']
        except KeyError:
            sim_params['horizon'] = 10
        try:
            sim_params['use_all_horizon'] = disc_params['use_all_horizon']
        except KeyError:
            sim_params['use_all_horizon'] = False
        try:
            sim_params['closed_loop'] = disc_params['closed_loop']
        except KeyError:
            sim_params['closed_loop'] = True
        if 'conservative' in disc_params:
            if disc_params['conservative'] is True:
                raise ValueError('MATLAB interface does not suport ' +
                                 'conservative simulation')
        output['simulation_parameters'] = sim_params

        # Abstraction
        output['abstraction'] = export_locations(abstraction)


    # Export the simulink model
    output['TS'] = export_mealy(mealy_machine, is_continuous)

    # Save file
    scipy.io.savemat(filename, output, oned_as='column')



def lti_export(ltisys):
    """Saves a LtiSysDyn as a Matlab struct."""

    output = {}
    output['A'] = ltisys.A
    output['B'] = ltisys.B
    output['E'] = ltisys.E
    output['K'] = ltisys.K
    output['domain'] = poly_export(ltisys.domain)
    output['Uset'] = poly_export(ltisys.Uset)
    output['Wset'] = poly_export(ltisys.Wset)

    return output


def pwa_export(pwasys):

    output = {}
    output['domain'] = poly_export(pwasys.domain)
    ltisystems = []
    for subsystem in pwasys.list_subsys:
        ltisystems.append(lti_export(subsystem))
    output['subsystems'] = ltisystems

    return output


def switched_export(switchedsys):

    output = {}
    output['disc_domain_size'] = list(switchedsys.disc_domain_size)
    output['cts_ss'] = poly_export(switchedsys.cts_ss)

    dynamics = []
    for label, system in switchedsys.dynamics.iteritems():
        system_dict = {}
        system_dict['env_act'] = label[0]
        system_dict['sys_act'] = label[1]
        system_dict['pwasys'] = pwa_export(system)
        dynamics.append(system_dict)

    output['dynamics'] = dynamics

    return output


def poly_export(poly):
    """Saves parts of a polytope as a dictionary for export to MATLAB.

    @param poly: L{Polytope} that will be exported.
    @return output: dictionary containing fields of poly
    """

    output = {}
    if poly is not None:
        output['A'] = poly.A
        output['b'] = poly.b
    return output


def reg_export(reg):
    """Saves a region as a dictionary for export to MATLAB.

    @type reg: L{Region}
    @return output: a dictionary containing a list of polytopes.
    """
    output = {}
    poly_list = []
    for poly in reg.list_poly:
        poly_list.append(poly_export(poly))
    output['list_poly'] = poly_list
    return output


def export_locations(abstraction):
    """Exports an abstraction to a .mat file

    @type abstraction: L{AbstractPwa} or L{AbstractSwitched}
    @rtype output: dictionary"""

    output = {}

    location_list = []
    for index, region in enumerate(abstraction.ppp.regions):
        location_dict = {}
        reg_dict = reg_export(region)
        location_dict['region'] = reg_dict
        location_dict['index'] = index
        location_list.append(location_dict)

    output['abstraction'] = location_list

    return output


def export_mealy_io(variables, values):
    """
    @rtype: list of dict
    """

    var_list = []
    for ind, variable in enumerate(variables):
        var_dict = {}
        var_dict['name'] = variable
        var_dict['values'] = list(values[ind])
        var_list.append(var_dict)
    return var_list


def export_mealy(mealy_machine, is_continuous):
    """Exports a Mealy Machine to data that can be put into a .mat file. Turns
    the Mealy Machine states into a list of dictionaries. Turns the Mealy
    Machine transitions into a list of transition matrices.

    Some of the exported information is a bit redundant, but makes writing the
    MATLAB code easier.

    @rtype: dict
    """

    output = {}

    # States will be exported as a list of dictionaries
    state_list = []
    for state_tuple in mealy_machine.states.find():

        # Do not export Sinit state
        if state_tuple[0] == 'Sinit':
            continue

        state_dict = {}
        state_dict['name'] = state_tuple[0]

        # For a continuous system, export the 'loc' variable
        if is_continuous:
            state_dict['loc'] = state_tuple[1]['loc']

        state_list.append(state_dict)
    output['states'] = state_list

    # Get list of environment and system variables
    env_vars = mealy_machine.inputs.keys()
    env_values = mealy_machine.inputs.values()
    sys_vars = mealy_machine.outputs.keys()
    sys_values = mealy_machine.outputs.values()
    #output['inputs'] = env_vars
    #output['outputs'] = sys_vars
    output['inputs'] = export_mealy_io(env_vars, env_values)
    output['outputs'] = export_mealy_io(sys_vars, sys_values)

    # Transitions will be exported as a 2D list of dictionaries. The only
    # purpose of this block here is to separate the inputs from the outputs to
    # make the MATLAB code easier to write
    transitions = []
    for transition_tuple in mealy_machine.transitions.find():

        # Ignore transitions from Sinit
        if transition_tuple[0] == 'Sinit':
            continue

        transition_vals = transition_tuple[2]
        transition_inputs = { var: str(transition_vals[var])
                              for var in env_vars }
        transition_outputs = { var: str(transition_vals[var])
                               for var in sys_vars }
        transition_dict = {}
        transition_dict['start_state'] = transition_tuple[0]
        transition_dict['end_state'] = transition_tuple[1]
        transition_dict['inputs'] = transition_inputs
        transition_dict['outputs'] = transition_outputs
        transitions.append(transition_dict)
    output['transitions'] = transitions


    # Initial states are the states that have transitions from Sinit. Initial
    # transitions (for the purposes of execution in Stateflow), are the
    # transitions coming from the states that transition from Sinit.
    Sinit_transitions = mealy_machine.transitions.find(from_states=['Sinit'])
    initial_states = [ trans[1] for trans in Sinit_transitions ]
    initial_transitions = mealy_machine.transitions.find(
        from_states=initial_states)
    initial_trans = []
    for init_transition in initial_transitions:

        transition_vals = init_transition[2]

        trans_dict = {}
        trans_dict['state'] = init_transition[1]
        trans_dict['inputs'] = { var: str(transition_vals[var])
                                 for var in env_vars }
        trans_dict['outputs'] = { var: str(transition_vals[var])
                                  for var in sys_vars }

        orig_state = init_transition[0]
        trans_dict['start_loc'] =  mealy_machine.states[orig_state]['loc']
        initial_trans.append(trans_dict)
    output['init_trans'] = initial_trans

    return output



"""
Tests for the abstraction from continuous dynamics to logic
"""
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
#logging.getLogger('tulip').setLevel(logging.ERROR)
logger.setLevel(logging.DEBUG)

from nose.tools import assert_raises

import matplotlib
# to avoid the need for using: ssh -X when running tests remotely
matplotlib.use('Agg')

import numpy as np

from tulip import abstract
from tulip.abstract import feasible
from tulip import hybrid
import polytope as pc

input_bound = 0.4

def subsys0():
    dom = pc.box2poly([[0., 3.], [0., 2.]])

    A = np.eye(2)
    B = np.eye(2)

    U = pc.box2poly([[0., 1.],
                  [0., 1.]])
    U.scale(input_bound)

    sys_dyn = hybrid.LtiSysDyn(A, B, Uset=U, domain=dom)

    return sys_dyn

def subsys1():
    dom = pc.box2poly([[0., 3.], [0., 2.]])

    A = np.eye(2)
    B = np.eye(2)

    U = pc.box2poly([[0., 0.],
                  [-1., 0.]])
    U.scale(input_bound)

    sys_dyn = hybrid.LtiSysDyn(A, B, Uset=U, domain=dom)

    return sys_dyn

def transition_directions_test():
    """
    unit test for correctness of abstracted transition directions, with:

      - uni-directional control authority
      - no disturbance
    """
    modes = []
    modes.append(('normal', 'fly'))
    modes.append(('refuel', 'fly'))
    env_modes, sys_modes = zip(*modes)

    cont_state_space = pc.box2poly([[0., 3.], [0., 2.]])
    pwa_sys = dict()
    pwa_sys[('normal', 'fly')] = hybrid.PwaSysDyn(
        [subsys0()], cont_state_space
    )
    pwa_sys[('refuel', 'fly')] = hybrid.PwaSysDyn(
        [subsys1()], cont_state_space
    )

    switched_dynamics = hybrid.SwitchedSysDyn(
        disc_domain_size=(len(env_modes), len(sys_modes)),
        dynamics=pwa_sys,
        env_labels=env_modes,
        disc_sys_labels=sys_modes,
        cts_ss=cont_state_space
    )

    cont_props = {}
    cont_props['home'] = pc.box2poly([[0., 1.], [0., 1.]])
    cont_props['lot'] = pc.box2poly([[2., 3.], [1., 2.]])

    ppp = abstract.prop2part(cont_state_space, cont_props)
    ppp, new2old = abstract.part2convex(ppp)

    N = 8
    trans_len=1

    disc_params = {}
    for mode in modes:
        disc_params[mode] = {'N':N, 'trans_length':trans_len}

    swab = abstract.discretize_switched(
        ppp, switched_dynamics, disc_params,
        plot=True, show_ts=True, only_adjacent=False
    )

    ts = swab.modes[('normal', 'fly')].ts
    edges = {(0, 0), (1, 1), (2, 2), (3, 3),
             (4, 4), (5, 5),
             (1, 2), (1, 4), (1, 5),
             (2, 3), (2, 5), (2, 0),
             (3, 0),
             (4, 5),
             (5, 0)}

    logger.debug(set(ts.edges() ).symmetric_difference(edges) )
    assert(set(ts.edges() ) == edges)

    ts = swab.ts

    assert(set(ts.edges() ) == edges)
    for i, j in edges:
        assert(ts[i][j][0]['env_actions'] == 'normal')
        assert(ts[i][j][0]['sys_actions'] == 'fly')

transition_directions_test.slow = True

def test_transient_regions():
    """drift is too strong, so no self-loop must exist

    This bug caused when running union is taken between Presets
    during solve_feasible, as happened with old use_all_horizon,
    cf:
        - 5b1e9681918739b276a221fcc1fd6eebfd058ce3
        - f5f4934ab9d21062f633eef3861ad935c3d3b54b
    """
    dom = pc.box2poly([[0.0, 4.0], [0.0, 3.0]])

    def cont_predicates():
        p = dict()
        p['safe'] = pc.box2poly([[0.5, 3.5], [0.5, 2.5]])

        ppp = abstract.prop2part(dom, p)
        ppp, new2old_reg = abstract.part2convex(ppp)
        ppp.plot()
        return ppp

    def drifting_dynamics():
        A = np.array([[1.0, 0.0],
                      [0.0, 1.0]])

        B = np.array([[1.0],
                      [0.0]])

        U = pc.box2poly([[0.0, 1.0]])

        K = np.array([[-100.0],
                      [0.0]])

        sys = hybrid.LtiSysDyn(A, B, None, K, U, None, dom)
        return sys

    ppp = cont_predicates()
    sys = drifting_dynamics()
    logger.info(sys)
    ab = abstract.discretize(ppp, sys, N=1, use_all_horizon=True,
                             trans_length=1)
    logger.debug(ab.ts)
    self_loops = {i for i,j in ab.ts.transitions() if i==j}
    logger.debug('self loops at states: ' + str(self_loops))

    assert(not self_loops)

    #ax = ab.plot(show_ts=True)
    #ax.figure.savefig('./very_simple.pdf')

def define_partition(dom):
    p = dict()
    p['a'] = pc.box2poly([[0.0, 10.0], [15.0, 18.0]])
    p['b'] = pc.box2poly([[0.0, 1.0], [0.0, 20.0]])

    ppp = abstract.prop2part(dom, p)
    ppp, new2old_reg = abstract.part2convex(ppp)
    return ppp

def define_dynamics(dom):
    A = np.eye(2)

    B = np.array([[1.0, -1.0],
                  [0.0, +1.0]])

    U = pc.box2poly([[0.0, 3.0],
                     [-3.0, 3.0]])

    E = np.array([[0.0],
                  [-1.0]])

    W = pc.box2poly([[-1.0, 1.0]])
    W.scale(0.4)

    K = np.array([[0.0],
                  [-0.4]])

    sys = hybrid.LtiSysDyn(A, B, E, K, U, W, dom)
    return sys

def test_abstract_the_dynamics():
    """test_abstract_the_dynamics (known to fail without GLPK)"""
    dom = pc.box2poly([[0.0, 10.0], [0.0, 20.0]])
    ppp = define_partition(dom)
    sys = define_dynamics(dom)
    logger.info(sys)

    disc_options = {'N':3, 'trans_length':2, 'min_cell_volume':1.5}

    ab = abstract.discretize(ppp, sys, plotit=False,
                             save_img=False, **disc_options)
    assert(ab.ppp.compute_adj() )
    assert(ab.ppp.is_partition() )
    #ax = ab.plot(show_ts=True, color_seed=0)
    #sys.plot(ax, show_domain=False)
    #print(ab.ts)

    #self_loops = {i for i,j in ab.ts.transitions() if i==j}
    #print('self loops at states: ' + str(self_loops))

test_abstract_the_dynamics.slow = True


def test_is_feasible():
    """Difference between attractor and fixed horizon."""
    dom = pc.box2poly([[0.0, 4.0], [0.0, 3.0]])
    sys = drifting_dynamics(dom)
    p1 = pc.box2poly([[0.0, 1.0], [0.0, 1.0]])
    p2 = pc.box2poly([[2.0, 3.0], [0.0, 1.0]])
    n = 10
    r = feasible.is_feasible(p1, p2, sys, n, use_all_horizon=False)
    assert r is False, r
    r = feasible.is_feasible(p1, p2, sys, n, use_all_horizon=True)
    assert r is True, r


def drifting_dynamics(dom):
    A = np.array([[1.0, 0.0],
                  [0.0, 1.0]])
    B = np.array([[1.0],
                  [0.0]])
    U = pc.box2poly([[0.0, 1.0]])
    K = np.array([[1.0],
                  [0.0]])
    sys = hybrid.LtiSysDyn(A, B, None, K, U, None, dom)
    return sys


if __name__ == '__main__':
    test_abstract_the_dynamics()



#!/usr/bin/env python
"""Discrete synthesis from a dummy abstraction of switched dynamics.

This is an example to demonstrate how the output of abstracting a switched
system, where the only control over the dynamics is through mode switching,
might look like.

We will assume, we have the 6 cell robot example.

     +---+---+---+
     | 3 | 4 | 5 |
     +---+---+---+
     | 0 | 1 | 2 |
     +---+---+---+
"""
# NO, 26 Jul 2013.
import numpy as np
from scipy import sparse as sp
from tulip import spec
from tulip import synth
from tulip import transys

###############################
# Switched system with 4 modes:
###############################

# In this scenario we have limited actions "left, right, up, down" with
# uncertain (nondeterministics) outcomes (e.g., due to bad actuators or
# bad low-level feedback controllers)

# Only control over the dynamics is through mode switching
# Transitions should be interpreted as nondeterministic

# Create a finite transition system
env_sws = transys.FTS()
env_sws.owner = 'env'

env_sws.sys_actions.add_from({'right','up','left','down'})

# str states
n = 6
states = transys.prepend_with(range(n), 's')
env_sws.states.add_from(set(states) )
env_sws.states.initial.add('s0')

env_sws.atomic_propositions.add_from(['home','lot'])
state_labels = [{'home'}, set(), set(), set(), set(), {'lot'}]

# Add states and decorate TS with state labels (aka atomic propositions)
for state, label in zip(states, state_labels):
    env_sws.states.add(state, ap=label)

# mode1 transitions
transmat1 = np.array([[0,1,0,0,1,0],
                      [0,0,1,0,0,1],
                      [0,0,1,0,0,0],
                      [0,1,0,0,1,0],
                      [0,0,1,0,0,1],
                      [0,0,0,0,0,1]])
env_sws.transitions.add_adj(
    sp.lil_matrix(transmat1), states, sys_actions='right'
)

# mode2 transitions
transmat2 = np.array([[0,0,0,1,0,0],
                      [0,0,0,0,1,1],
                      [0,0,0,0,0,1],
                      [0,0,0,1,0,0],
                      [0,0,0,0,1,0],
                      [0,0,0,0,0,1]])
env_sws.transitions.add_adj(
    sp.lil_matrix(transmat2), states, sys_actions='up'
)

# mode3 transitions
transmat3 = np.array([[1,0,0,0,0,0],
                      [1,0,0,1,0,0],
                      [0,1,0,0,1,0],
                      [0,0,0,1,0,0],
                      [1,0,0,1,0,0],
                      [0,1,0,0,1,0]])
env_sws.transitions.add_adj(
    sp.lil_matrix(transmat3), states, sys_actions='left'
)

# mode4 transitions
transmat4 = np.array([[1,0,0,0,0,0],
                      [0,1,0,0,0,0],
                      [0,0,1,0,0,0],
                      [1,0,0,0,0,0],
                      [0,1,1,0,0,0],
                      [0,0,1,0,0,0]])
env_sws.transitions.add_adj(
    sp.lil_matrix(transmat4), states, sys_actions='down'
)

# This is what is visible to the outside world (and will go into synthesis method)
print(env_sws)

#
# Environment variables and specification
#
# The environment can issue a park signal that the robot just respond
# to by moving to the lower left corner of the grid.  We assume that
# the park signal is turned off infinitely often.
#
env_vars = {'park'}
env_init = {'park'}
env_prog = {'!park'}
env_safe = set()                # empty set

#
# System specification
#
# The system specification is that the robot should repeatedly revisit
# the upper right corner of the grid while at the same time responding
# to the park signal by visiting the lower left corner.  The LTL
# specification is given by
#
#     []<> home && [](park -> <>lot)
#
# Since this specification is not in GR(1) form, we introduce the
# variable X0reach that is initialized to True and the specification
# [](park -> <>lot) becomes
#
#     [](X (X0reach) <-> lot || (X0reach && !park))
#

# Augment the environmental description to make it GR(1)
#! TODO: create a function to convert this type of spec automatically

# Define the specification
#! NOTE: maybe "synthesize" should infer the atomic proposition from the
# transition system? Or, we can declare the mode variable, and the values
# of the mode variable are read from the transition system.
sys_vars = {'X0reach'}
sys_init = {'X0reach','sys_actions = "right"'}
sys_prog = {'home'}               # []<>home
sys_safe = {'(X (X0reach) <-> lot) || (X0reach && !park)'}
sys_prog |= {'X0reach'}

# Create the specification
specs = spec.GRSpec(env_vars, sys_vars, env_init, sys_init,
                    env_safe, sys_safe, env_prog, sys_prog)
specs.qinit = '\E \A'

# Controller synthesis
#
# At this point we can synthesize the controller using one of the available
# methods.
#
ctrl = synth.synthesize('omega', specs, env=env_sws)
assert ctrl is not None, 'unrealizable'

# Generate a graphical representation of the controller for viewing
if not ctrl.save('only_mode_controlled.png'):
    print(ctrl)
